"""
    TXT/IGOR 2D Q Map file reader
"""
#####################################################################
#This software was developed by the University of Tennessee as part of the
#Distributed Data Analysis of Neutron Scattering Experiments (DANSE)
#project funded by the US National Science Foundation.
#See the license text in license.txt
#copyright 2008, University of Tennessee
######################################################################
import os
import time
from typing import Any

import numpy as np

from sasdata.data_util.nxsunit import Converter

from sasdata.dataloader.data_info import plottable_2D, Data2D, DataInfo, Detector
from sasdata.dataloader.filereader import FileReader
from sasdata.data_util.loader_exceptions import FileContentsException


def check_point(x_point: Any):
    """
    check point validity
    """
    # set zero for non_floats
    try:
        return float(x_point)
    except Exception:
        return 0


class Reader(FileReader):
    """ Simple data reader for Igor data files """
    # File type
    type_name = "IGOR/DAT 2D Q_map"
    # Wildcards
    type = ["IGOR/DAT 2D file in Q_map (*.dat)|*.DAT"]
    # Extension
    ext = ['.DAT', '.dat']

    def write(self, filename: str, data: Data2D):
        """
        Write to .dat

        :param filename: file name to write
        :param data: data2D
        """
        # Write the file
        try:
            fd = open(filename, 'w')
            t = time.localtime()
            time_str = time.strftime("%H:%M on %b %d %y", t)

            header_str = "Data columns are Qx - Qy - I(Qx,Qy)\n\nASCII data"
            header_str += " created at %s \n\n" % time_str
            # simple 2D header
            fd.write(header_str)
            # write qx qy I values
            for i in range(len(data.data)):
                fd.write("%g  %g  %g\n" % (data.qx_data[i],
                                            data.qy_data[i],
                                           data.data[i]))
        finally:
            if fd and not fd.closed:
                fd.close()

    def get_file_contents(self):
        # Read file
        buf = self.readall()
        self.f_open.close()
        # Instantiate data object
        self.current_dataset = plottable_2D()
        self.current_datainfo = DataInfo()
        self.current_datainfo.filename = os.path.basename(self.f_open.name)
        self.current_datainfo.detector.append(Detector())

        # Get content
        data_started = False

        # Defaults
        lines = buf.split('\n')
        x = []
        y = []

        wavelength = None
        distance = None
        transmission = None

        pixel_x = None
        pixel_y = None

        is_info = False
        is_center = False

        # Remove the last lines before the for loop if the lines are empty
        # to calculate the exact number of data points
        count = 0
        while (len(lines[len(lines) - (count + 1)].lstrip().rstrip()) < 1):
            del lines[len(lines) - (count + 1)]
            count = count + 1

        # Read Header and find the dimensions of 2D data
        line_num = 0
        # Old version NIST files: 0
        ver = 0
        for line in lines:
            line_num += 1
            # Reading the header applies only to IGOR/NIST 2D q_map data files
            # Find setup info line
            if is_info:
                is_info = False
                line_toks = line.split()
                # Wavelength in Angstrom
                try:
                    wavelength = float(line_toks[1])
                    # Wavelength is stored in angstroms; convert if necessary
                    if self.current_datainfo.source.wavelength_unit != 'A':
                        conv = Converter('A')
                        wavelength = conv(wavelength,
                                          units=self.current_datainfo.source.wavelength_unit)
                except Exception:
                    pass  # Not required
                try:
                    distance = float(line_toks[3])
                    # Distance is stored in meters; convert if necessary
                    if self.current_datainfo.detector[0].distance_unit != 'm':
                        conv = Converter('m')
                        distance = conv(distance,
                            units=self.current_datainfo.detector[0].distance_unit)
                except Exception:
                    pass  # Not required

                try:
                    transmission = float(line_toks[4])
                except Exception:
                    pass  # Not required

            if line.count("LAMBDA") > 0:
                is_info = True

            # Find center info line
            if is_center:
                is_center = False
                line_toks = line.split()
                # Center in bin number
                center_x = float(line_toks[0])
                center_y = float(line_toks[1])

            if line.count("BCENT") > 0:
                is_center = True
            # Check version
            if line.count("Data columns") > 0:
                if line.count("err(I)") > 0:
                    ver = 1
            # Find data start
            if line.count("ASCII data") > 0:
                data_started = True
                continue

            # Read and get data.
            if data_started:
                line_toks = line.split()
                if len(line_toks) == 0:
                    #empty line
                    continue
                # the number of columns must be stayed same
                col_num = len(line_toks)
                break

        # Make numpy array to remove header lines using index
        lines_array = np.array(lines)

        # index for lines_array
        lines_index = np.arange(len(lines))

        # get the data lines
        data_lines = lines_array[lines_index >= (line_num - 1)]
        # Now we get the total number of rows (i.e., # of data points)
        row_num = len(data_lines)
        # make it as list again to control the separators
        data_list = " ".join(data_lines.tolist())
        # split all data to one big list w/" "separator
        data_list = data_list.split()

        # Check if the size is consistent with data, otherwise
        # try the tab(\t) separator
        # (this may be removed once get the confidence
        # the former working all cases).
        if len(data_list) != (len(data_lines)) * col_num:
            data_list = "\t".join(data_lines.tolist())
            data_list = data_list.split()

        # Change it(string) into float
        # data_list = map(float,data_list)
        data_list1 = list(map(check_point, data_list))

        # numpy array form
        data_array = np.array(data_list1)
        # Redimesion based on the row_num and col_num,
        # otherwise raise an error.
        try:
            data_point = data_array.reshape(row_num, col_num).transpose()
        except Exception:
            msg = "red2d_reader can't read this file: Incorrect number of data points provided."
            raise FileContentsException(msg)
        # Get the all data: Let's HARDcoding; Todo find better way
        # Defaults
        dqx_data = np.zeros(0)
        dqy_data = np.zeros(0)
        err_data = np.ones(row_num)
        qz_data = np.zeros(row_num)
        mask = np.ones(row_num, dtype=bool)
        # Get from the array
        qx_data = data_point[0]
        qy_data = data_point[1]
        data = data_point[2]
        if ver == 1:
            if col_num > (2 + ver):
                err_data = data_point[(2 + ver)]
        if col_num > (3 + ver):
            qz_data = data_point[(3 + ver)]
        if col_num > (4 + ver):
            dqx_data = data_point[(4 + ver)]
        if col_num > (5 + ver):
            dqy_data = data_point[(5 + ver)]
        # Column '6 + ver' is the shadow factor value. A separate mask column
        #   was added to account for self-drawn masks.
        # if col_num > (6 + ver): mask[data_point[(6 + ver)] < 1] = False
        if col_num > (7 + ver):
            mask = np.invert(np.asarray(data_point[(7 + ver)], dtype=bool))
        q_data = np.sqrt(qx_data*qx_data+qy_data*qy_data+qz_data*qz_data)

        # Store limits of the image in q space
        xmin = np.min(qx_data)
        xmax = np.max(qx_data)
        ymin = np.min(qy_data)
        ymax = np.max(qy_data)

        # Find unique Qx and Qy values for data binning and visualization
        # len(x_bins) * len(y_bins) ~= len(qx_data) ~= len(qy_data)
        x_bins = np.unique(qx_data)
        y_bins = np.unique(qy_data)
        # For non-uniform qx_data and/or qy_data
        #  Cases: Rotated detectors, floating point variations
        if round(len(x_bins) * len(y_bins) / len(qx_data)) >= 2:
            # qx_data increases along rows => travel along a single pixel line
            num_qx = np.argmax(np.hstack((qx_data[1:] < qx_data[:-1], True)))
            x_bins = qx_data[:num_qx + 1]
            # qy_data increases along columns => transpose qx_data shape
            qy = np.reshape(qy_data, (len(qx_data)//len(x_bins), len(x_bins)))
            y_bins = np.transpose(qy)[0].tolist()

        # Store data in outputs
        self.current_dataset.data = data
        if (err_data == 1).all():
            self.current_dataset.err_data = np.sqrt(np.abs(data))
            self.current_dataset.err_data[self.current_dataset.err_data == 0.0] = 1.0
        else:
            self.current_dataset.err_data = err_data

        self.current_dataset.qx_data = qx_data
        self.current_dataset.qy_data = qy_data
        self.current_dataset.q_data = q_data
        self.current_dataset.mask = mask

        self.current_dataset.x_bins = x_bins
        self.current_dataset.y_bins = y_bins

        self.current_dataset.xmin = xmin
        self.current_dataset.xmax = xmax
        self.current_dataset.ymin = ymin
        self.current_dataset.ymax = ymax

        self.current_datainfo.source.wavelength = wavelength

        # Store pixel size in mm
        self.current_datainfo.detector[0].pixel_size.x = pixel_x
        self.current_datainfo.detector[0].pixel_size.y = pixel_y

        # Store the sample to detector distance
        self.current_datainfo.detector[0].distance = distance

        # optional data: if all of dq data == 0, do not pass to output
        if len(dqx_data) == len(qx_data) and dqx_data.any() != 0:
            # if no dqx_data, do not pass dqy_data.
            # (1 axis dq is not supported yet).
            if len(dqy_data) == len(qy_data) and dqy_data.any() != 0:
                # Currently we do not support dq parr, perp.
                # transfer the comp. to cartesian coord. for newer version.
                if ver != 1:
                    diag = np.sqrt(qx_data * qx_data + qy_data * qy_data)
                    cos_th = qx_data / diag
                    sin_th = qy_data / diag
                    self.current_dataset.dqx_data = np.sqrt(
                        (dqx_data * cos_th)**2 + (dqy_data * sin_th)**2)
                    self.current_dataset.dqy_data = np.sqrt(
                        (dqx_data * sin_th)**2 + (dqy_data * cos_th)**2)
                else:
                    self.current_dataset.dqx_data = dqx_data
                    self.current_dataset.dqy_data = dqy_data

        # Units of axes
        self.current_dataset = self.set_default_2d_units(self.current_dataset)

        # Store loading process information
        self.current_datainfo.meta_data['loader'] = self.type_name

        self.send_to_output()



from sasdata.dataloader.loader import Loader
import unittest


class testLoader(unittest.TestCase):
    def setUp(self):
        self.s = Loader()

    def test_load_unknown(self):
        """
            self.s.load('blah.mat') on an unknown type
            should raise a ValueError exception (thrown by data_util.registry)
        """
        self.assertRaises(ValueError, self.s.load, 'angles_flat.mat')
        
    def test_corrupt(self):
        """
            Loading a corrupted file with a known extension
            should raise a runtime exception.
            The error condition is similar to an unknown 
            type (file extension). When a reader is identified as
            knowing the file extension, it tries to load. If it
            raises an exception, the system should try to identify
            another reader that knows about that file extension.
            If they all fail, we raise a runtime exception stating
            that no reader was found for this file.
        """
        self.assertRaises(RuntimeError, self.s.load, 'corrupt.png')


if __name__ == '__main__':
    unittest.main()



r"""
Uncertainty propagation class for arithmetic, log and exp.

Based on scalars or numpy vectors, this class allows you to store and
manipulate values+uncertainties, with propagation of gaussian error for
addition, subtraction, multiplication, division, power, exp and log.

Storage properties are determined by the numbers used to set the value
and uncertainty.  Be sure to use floating point uncertainty vectors
for inplace operations since numpy does not do automatic type conversion.
Normal operations can use mixed integer and floating point.  In place
operations such as *a \*= b* create at most one extra copy for each operation.
By contrast, *c = a\*b* uses four intermediate vectors, so shouldn't be used
for huge arrays.
"""

import numpy as np

from . import err1d
from .formatnum import format_uncertainty

__all__ = ['Uncertainty']

# TODO: rename to Measurement and add support for units?
# TODO: C implementation of *,/,**?
class Uncertainty(object):
    # Make standard deviation available
    def _getdx(self): return np.sqrt(self.variance)
    def _setdx(self,dx):
        # Direct operation
        #    variance = dx**2
        # Indirect operation to avoid temporaries
        self.variance[:] = dx
        self.variance **= 2
    dx = property(_getdx,_setdx,doc="standard deviation")

    # Constructor
    def __init__(self, x, variance=None):
        self.x, self.variance = x, variance

    # Numpy array slicing operations
    def __len__(self):
        return len(self.x)
    def __getitem__(self,key):
        return Uncertainty(self.x[key],self.variance[key])
    def __setitem__(self,key,value):
        self.x[key] = value.x
        self.variance[key] = value.variance
    def __delitem__(self, key):
        del self.x[key]
        del self.variance[key]
    #def __iter__(self): pass # Not sure we need iter

    # Normal operations: may be of mixed type
    def __add__(self, other):
        if isinstance(other,Uncertainty):
            return Uncertainty(*err1d.add(self.x,self.variance,other.x,other.variance))
        else:
            return Uncertainty(self.x+other, self.variance+0) # Force copy
    def __sub__(self, other):
        if isinstance(other,Uncertainty):
            return Uncertainty(*err1d.sub(self.x,self.variance,other.x,other.variance))
        else:
            return Uncertainty(self.x-other, self.variance+0) # Force copy
    def __mul__(self, other):
        if isinstance(other,Uncertainty):
            return Uncertainty(*err1d.mul(self.x,self.variance,other.x,other.variance))
        else:
            return Uncertainty(self.x*other, self.variance*other**2)
    def __truediv__(self, other):
        if isinstance(other,Uncertainty):
            return Uncertainty(*err1d.div(self.x,self.variance,other.x,other.variance))
        else:
            return Uncertainty(self.x/other, self.variance/other**2)
    def __pow__(self, other):
        if isinstance(other,Uncertainty):
            # Haven't calcuated variance in (a+/-da) ** (b+/-db)
            return NotImplemented
        else:
            return Uncertainty(*err1d.pow(self.x,self.variance,other))

    # Reverse operations
    def __radd__(self, other):
        return Uncertainty(self.x+other, self.variance+0) # Force copy
    def __rsub__(self, other):
        return Uncertainty(other-self.x, self.variance+0)
    def __rmul__(self, other):
        return Uncertainty(self.x*other, self.variance*other**2)
    def __rtruediv__(self, other):
        x,variance = err1d.pow(self.x,self.variance,-1)
        return Uncertainty(x*other,variance*other**2)
    def __rpow__(self, other): return NotImplemented

    # In-place operations: may be of mixed type
    def __iadd__(self, other):
        if isinstance(other,Uncertainty):
            self.x,self.variance \
                = err1d.add_inplace(self.x,self.variance,other.x,other.variance)
        else:
            self.x+=other
        return self
    def __isub__(self, other):
        if isinstance(other,Uncertainty):
            self.x,self.variance \
                = err1d.sub_inplace(self.x,self.variance,other.x,other.variance)
        else:
            self.x-=other
        return self
    def __imul__(self, other):
        if isinstance(other,Uncertainty):
            self.x, self.variance \
                = err1d.mul_inplace(self.x,self.variance,other.x,other.variance)
        else:
            self.x *= other
            self.variance *= other**2
        return self
    def __itruediv__(self, other):
        if isinstance(other,Uncertainty):
            self.x,self.variance \
                = err1d.div_inplace(self.x,self.variance,other.x,other.variance)
        else:
            self.x /= other
            self.variance /= other**2
        return self
    def __ipow__(self, other):
        if isinstance(other,Uncertainty):
            # Haven't calcuated variance in (a+/-da) ** (b+/-db)
            return NotImplemented
        else:
            self.x,self.variance = err1d.pow_inplace(self.x, self.variance, other)
        return self

    # Use true division instead of integer division
    def __div__(self, other): return self.__truediv__(other)
    def __rdiv__(self, other): return self.__rtruediv__(other)
    def __idiv__(self, other): return self.__itruediv__(other)


    # Unary ops
    def __neg__(self):
        return Uncertainty(-self.x,self.variance)
    def __pos__(self):
        return self
    def __abs__(self):
        return Uncertainty(np.abs(self.x),self.variance)

    def __str__(self):
        #return str(self.x)+" +/- "+str(np.sqrt(self.variance))
        if np.isscalar(self.x):
            return format_uncertainty(self.x,np.sqrt(self.variance))
        else:
            return [format_uncertainty(v,dv)
                    for v,dv in zip(self.x,np.sqrt(self.variance))]
    def __repr__(self):
        return "Uncertainty(%s,%s)"%(str(self.x),str(self.variance))

    # Not implemented
    def __floordiv__(self, other): return NotImplemented
    def __mod__(self, other): return NotImplemented
    def __divmod__(self, other): return NotImplemented
    def __mod__(self, other): return NotImplemented
    def __lshift__(self, other): return NotImplemented
    def __rshift__(self, other): return NotImplemented
    def __and__(self, other): return NotImplemented
    def __xor__(self, other): return NotImplemented
    def __or__(self, other): return NotImplemented

    def __rfloordiv__(self, other): return NotImplemented
    def __rmod__(self, other): return NotImplemented
    def __rdivmod__(self, other): return NotImplemented
    def __rmod__(self, other): return NotImplemented
    def __rlshift__(self, other): return NotImplemented
    def __rrshift__(self, other): return NotImplemented
    def __rand__(self, other): return NotImplemented
    def __rxor__(self, other): return NotImplemented
    def __ror__(self, other): return NotImplemented

    def __ifloordiv__(self, other): return NotImplemented
    def __imod__(self, other): return NotImplemented
    def __idivmod__(self, other): return NotImplemented
    def __imod__(self, other): return NotImplemented
    def __ilshift__(self, other): return NotImplemented
    def __irshift__(self, other): return NotImplemented
    def __iand__(self, other): return NotImplemented
    def __ixor__(self, other): return NotImplemented
    def __ior__(self, other): return NotImplemented

    def __invert__(self): return NotImplmented  # For ~x
    def __complex__(self): return NotImplmented
    def __int__(self): return NotImplmented
    def __long__(self): return NotImplmented
    def __float__(self): return NotImplmented
    def __oct__(self): return NotImplmented
    def __hex__(self): return NotImplmented
    def __index__(self): return NotImplmented
    def __coerce__(self): return NotImplmented

    def log(self):
        return Uncertainty(*err1d.log(self.x,self.variance))

    def exp(self):
        return Uncertainty(*err1d.exp(self.x,self.variance))

def log(val): return self.log()
def exp(val): return self.exp()

def test():
    a = Uncertainty(5,3)
    b = Uncertainty(4,2)

    # Scalar operations
    z = a+4
    assert z.x == 5+4 and z.variance == 3
    z = a-4
    assert z.x == 5-4 and z.variance == 3
    z = a*4
    assert z.x == 5*4 and z.variance == 3*4**2
    z = a/4
    assert z.x == 5./4 and z.variance == 3./4**2

    # Reverse scalar operations
    z = 4+a
    assert z.x == 4+5 and z.variance == 3
    z = 4-a
    assert z.x == 4-5 and z.variance == 3
    z = 4*a
    assert z.x == 4*5 and z.variance == 3*4**2
    z = 4/a
    assert z.x == 4./5 and abs(z.variance - 3./5**4 * 4**2) < 1e-15

    # Power operations
    z = a**2
    assert z.x == 5**2 and z.variance == 4*3*5**2
    z = a**1
    assert z.x == 5**1 and z.variance == 3
    z = a**0
    assert z.x == 5**0 and z.variance == 0
    z = a**-1
    assert z.x == 5**-1 and abs(z.variance - 3./5**4) < 1e-15

    # Binary operations
    z = a+b
    assert z.x == 5+4 and z.variance == 3+2
    z = a-b
    assert z.x == 5-4 and z.variance == 3+2
    z = a*b
    assert z.x == 5*4 and z.variance == (5**2*2 + 4**2*3)
    z = a/b
    assert z.x == 5./4 and abs(z.variance - (3./5**2 + 2./4**2)*(5./4)**2) < 1e-15

    # ===== Inplace operations =====
    # Scalar operations
    y = a+0; y += 4
    z = a+4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y -= 4
    z = a-4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y *= 4
    z = a*4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y /= 4
    z = a/4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15

    # Power operations
    y = a+0; y **= 4
    z = a**4
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15

    # Binary operations
    y = a+0; y += b
    z = a+b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y -= b
    z = a-b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y *= b
    z = a*b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15
    y = a+0; y /= b
    z = a/b
    assert y.x == z.x and abs(y.variance-z.variance) < 1e-15


    # =============== vector operations ================
    # Slicing
    z = Uncertainty(np.array([1,2,3,4,5]),np.array([2,1,2,3,2]))
    assert z[2].x == 3 and z[2].variance == 2
    assert (z[2:4].x == [3,4]).all()
    assert (z[2:4].variance == [2,3]).all()
    z[2:4] = Uncertainty(np.array([8,7]),np.array([4,5]))
    assert z[2].x == 8 and z[2].variance == 4
    A = Uncertainty(np.array([a.x]*2),np.array([a.variance]*2))
    B = Uncertainty(np.array([b.x]*2),np.array([b.variance]*2))

    # TODO complete tests of copy and inplace operations for vectors and slices.

    # Binary operations
    z = A+B
    assert (z.x == 5+4).all() and (z.variance == 3+2).all()
    z = A-B
    assert (z.x == 5-4).all() and (z.variance == 3+2).all()
    z = A*B
    assert (z.x == 5*4).all() and (z.variance == (5**2*2 + 4**2*3)).all()
    z = A/B
    assert (z.x == 5./4).all()
    assert (abs(z.variance - (3./5**2 + 2./4**2)*(5./4)**2) < 1e-15).all()

    # printing; note that sqrt(3) ~ 1.7
    assert str(Uncertainty(5,3)) == "5.0(17)"
    assert str(Uncertainty(15,3)) == "15.0(17)"
    assert str(Uncertainty(151.23356,0.324185**2)) == "151.23(32)"

if __name__ == "__main__": test()



"""
    Generic XML read and write utility

    Usage: Either extend xml_reader or add as a class variable.
"""
############################################################################
#This software was developed by the University of Tennessee as part of the
#Distributed Data Analysis of Neutron Scattering Experiments (DANSE)
#project funded by the US National Science Foundation.
#If you use DANSE applications to do scientific research that leads to
#publication, we ask that you acknowledge the use of the software with the
#following sentence:
#This work benefited from DANSE software developed under NSF award DMR-0520547.
#copyright 2008,2009 University of Tennessee
#############################################################################

import logging

from lxml import etree
from lxml.builder import E

from sasdata.dataloader.filereader import FileReader, decode

logger = logging.getLogger(__name__)

PARSER = etree.ETCompatXMLParser(remove_comments=True, remove_pis=False)


class XMLreader(FileReader):
    """
    Generic XML read and write class. Mostly helper functions.
    Makes reading/writing XML a bit easier than calling lxml libraries directly.

    :Dependencies:
        This class requires lxml 2.3 or higher.
    """

    xml = None
    xmldoc = None
    xmlroot = None
    schema = None
    schemadoc = None
    encoding = None
    processing_instructions = None

    def __init__(self, xml=None, schema=None):
        super(XMLreader, self).__init__()
        self.xml = xml
        self.schema = schema
        self.processing_instructions = {}
        if xml is not None:
            self.set_xml_file(xml)
        else:
            self.xmldoc = None
            self.xmlroot = None
        if schema is not None:
            self.set_schema(schema)
        else:
            self.schemadoc = None

    def reader(self):
        """
        Read in an XML file into memory and return an lxml dictionary
        """
        if self.validate_xml():
            self.xmldoc = etree.parse(self.xml, parser=PARSER)
        else:
            raise etree.XMLSchemaValidateError(self, self.find_invalid_xml())
        return self.xmldoc

    def set_xml_file(self, xml):
        """
        Set the XML file and parse
        """
        try:
            self.xml = xml
            self.xmldoc = etree.parse(self.xml, parser=PARSER)
            self.xmlroot = self.xmldoc.getroot()
        except etree.XMLSyntaxError as xml_error:
            logger.info(xml_error)
            raise xml_error
        except Exception:
            self.xml = None
            self.xmldoc = None
            self.xmlroot = None

    def set_xml_string(self, tag_soup):
        """
        Set an XML string as the working XML.

        :param tag_soup: XML formatted string
        """
        try:
            self.xml = tag_soup
            self.xmldoc = tag_soup
            self.xmlroot = etree.fromstring(tag_soup)
        except etree.XMLSyntaxError as xml_error:
            logger.info(xml_error)
            raise xml_error
        except Exception as exc:
            self.xml = None
            self.xmldoc = None
            self.xmlroot = None
            raise exc

    def set_schema(self, schema):
        """
        Set the schema file and parse
        """
        try:
            self.schema = schema
            self.schemadoc = etree.parse(self.schema, parser=PARSER)
        except etree.XMLSyntaxError as xml_error:
            logger.info(xml_error)
        except Exception:
            self.schema = None
            self.schemadoc = None

    def validate_xml(self):
        """
        Checks to see if the XML file meets the schema
        """
        valid = True
        if self.schema is not None:
            self.parse_schema_and_doc()
            schema_check = etree.XMLSchema(self.schemadoc)
            valid = schema_check.validate(self.xmldoc)
        return valid

    def find_invalid_xml(self):
        """
        Finds the first offending element that should not be present in XML file
        """
        first_error = ""
        self.parse_schema_and_doc()
        schema = etree.XMLSchema(self.schemadoc)
        try:
            first_error = schema.assertValid(self.xmldoc)
        except etree.DocumentInvalid as err:
            # Suppress errors for <'any'> elements
            if "##other" in str(err):
                return first_error
            first_error = str(err)
        return first_error

    def parse_schema_and_doc(self):
        """
        Creates a dictionary of the parsed schema and xml files.
        """
        self.set_xml_file(self.xml)
        self.set_schema(self.schema)

    def to_string(self, elem, pretty_print=False, encoding=None):
        """
        Converts an etree element into a string
        """
        return decode(etree.tostring(elem, pretty_print=pretty_print,
                                     encoding=encoding))

    def break_processing_instructions(self, string, dic):
        """
        Method to break a processing instruction string apart and add to a dict

        :param string: A processing instruction as a string
        :param dic: The dictionary to save the PIs to
        """
        pi_string = string.replace("<?", "").replace("?>", "")
        split = pi_string.split(" ", 1)
        pi_name = split[0]
        attr = split[1]
        new_pi_name = self._create_unique_key(dic, pi_name)
        dic[new_pi_name] = attr
        return dic

    def set_processing_instructions(self):
        """
        Take out all processing instructions and create a dictionary from them
        If there is a default encoding, the value is also saved
        """
        dic = {}
        proc_instr = self.xmlroot.getprevious()
        while proc_instr is not None:
            pi_string = self.to_string(proc_instr)
            if "?>\n<?" in pi_string:
                pi_string = pi_string.split("?>\n<?")
            if isinstance(pi_string, str):
                dic = self.break_processing_instructions(pi_string, dic)
            elif isinstance(pi_string, list):
                for item in pi_string:
                    dic = self.break_processing_instructions(item, dic)
            proc_instr = proc_instr.getprevious()
        if 'xml' in dic:
            self.set_encoding(dic['xml'])
            del dic['xml']
        self.processing_instructions = dic

    def set_encoding(self, attr_str):
        """
        Find the encoding in the xml declaration and save it as a string

        :param attr_str: All attributes as a string
            e.g. "foo1="bar1" foo2="bar2" foo3="bar3" ... foo_n="bar_n""
        """
        attr_str = attr_str.replace(" = ", "=")
        attr_list = attr_str.split()
        for item in attr_list:
            name_value = item.split("\"=")
            name = name_value[0].lower()
            value = name_value[1]
            if name == "encoding":
                self.encoding = value
                return
        self.encoding = None

    def _create_unique_key(self, dictionary, name, numb=0):
        """
        Create a unique key value for any dictionary to prevent overwriting
        Recurses until a unique key value is found.

        :param dictionary: A dictionary with any number of entries
        :param name: The index of the item to be added to dictionary
        :param numb: The number to be appended to the name, starts at 0
        """
        if dictionary.get(name) is not None:
            numb += 1
            name = name.split("_")[0]
            name += "_{0}".format(numb)
            name = self._create_unique_key(dictionary, name, numb)
        return name

    def create_tree(self, root):
        """
        Create an element tree for processing from an etree element

        :param root: etree Element(s)
        """
        return etree.ElementTree(root)

    def create_element_from_string(self, xml_string):
        """
        Create an element from an XML string

        :param xml_string: A string of xml
        """
        return etree.fromstring(xml_string)

    def create_element(self, name, attrib=None, nsmap=None):
        """
        Create an XML element for writing to file

        :param name: The name of the element to be created
        """
        if attrib is None:
            attrib = {}
        return etree.Element(name, attrib, nsmap)

    def write_text(self, elem, text):
        """
        Write text to an etree Element

        :param elem: etree.Element object
        :param text: text to write to the element
        """
        elem.text = text
        return elem

    def write_attribute(self, elem, attr_name, attr_value):
        """
        Write attributes to an Element

        :param elem: etree.Element object
        :param attr_name: attribute name to write
        :param attr_value: attribute value to set
        """
        attr = elem.attrib
        attr[attr_name] = attr_value

    def return_processing_instructions(self):
        """
        Get all processing instructions saved when loading the document

        :param tree: etree.ElementTree object to write PIs to
        """
        pi_list = []
        if self.processing_instructions is not None:
            for key in self.processing_instructions:
                value = self.processing_instructions.get(key)
                pi_item = etree.ProcessingInstruction(key, value)
                pi_list.append(pi_item)
        return pi_list

    def append(self, element, tree):
        """
        Append an etree Element to an ElementTree.

        :param element: etree Element to append
        :param tree: ElementTree object to append to
        """
        tree = tree.append(element)
        return tree

    def ebuilder(self, parent, elementname, text=None, attrib=None):
        """
        Use lxml E builder class with arbitrary inputs.

        :param parnet: The parent element to append a child to
        :param elementname: The name of the child in string form
        :param text: The element text
        :param attrib: A dictionary of attribute names to attribute values
        """
        text = str(text)
        if attrib is None:
            attrib = {}
        elem = E(elementname, attrib, text)
        parent = parent.append(elem)
        return parent



"""
Information relating to the CanSAS data format. These constants are used in
the cansas_reader.py file to read in any version of the cansas format.
"""


class CansasConstants:
    """
    The base class to define where all of the data is to be saved by cansas_reader.py.
    """
    names = ''
    format = ''

    def __init__(self):
        self.names = self.CANSAS_NS
        self.format = self.CANSAS_FORMAT

    def iterate_namespace(self, namespace):
        """
        Method to iterate through a cansas constants tree based on a list of
        names

        :param namespace: A list of names that match the tree structure of
            cansas_constants
        """
        # The current level to look through in cansas_constants.
        return_me = CurrentLevel()
        return_me.current_level = self.CANSAS_FORMAT.get("SASentry")
        # Defaults for variable and datatype
        return_me.ns_datatype = "content"
        return_me.ns_optional = True
        for name in namespace:
            try:
                if name != "SASentry":
                    return_me.current_level = \
                        return_me.current_level.get("children").get(name, "")
                    if return_me.current_level == "":
                        return_me.current_level = \
                                return_me.current_level.get("<any>", "")
                    cl_datatype = return_me.current_level.get("storeas", "")
                    cl_units_optional = \
                             return_me.current_level.get("units_optional", "")
                    # Where are how to store the variable for the given
                    # namespace CANSAS_CONSTANTS tree is hierarchical, so
                    # is no value, inherit
                    return_me.ns_datatype = cl_datatype if cl_datatype != "" \
                        else return_me.ns_datatype
                    return_me.ns_optional = cl_units_optional if \
                        cl_units_optional != return_me.ns_optional \
                                        else return_me.ns_optional
            except AttributeError:
                return_me.ns_datatype = "content"
                return_me.ns_optional = True
        return return_me

    def get_namespace_map(self):
        """
        Helper method to get the names namespace list
        """
        return self.names

    # CANSAS_NS holds the base namespace and default schema file information
    CANSAS_NS = {"1.0" : {"ns" : "cansas1d/1.0",
                          "schema" : "cansas1d_v1_0.xsd"
                         },
                 "1.1" : {"ns" : "urn:cansas1d:1.1",
                          "schema" : "cansas1d_v1_1.xsd"
                         }
                }

    # The constants below hold information on where to store the CanSAS data
    # when loaded in using sasview
    ANY = {"storeas" : "content"}
    TITLE = {}
    SASNOTE = {}
    SASPROCESS_TERM = {"attributes" : {"unit" : {}, "name" : {}}}
    SASPROCESS_SASPROCESSNOTE = {"children" : {"<any>" : ANY}}
    SASPROCESS = {"children" : {"name" : {},
                                "date" : {},
                                "description" : {},
                                "term" : SASPROCESS_TERM,
                                "SASprocessnote" : SASPROCESS_SASPROCESSNOTE,
                                "<any>" : ANY
                               },
                 }
    RUN = {"attributes" : {"name" :{}}}
    SASDATA_IDATA_Q = {"units_optional" : False,
                       "storeas" : "float",
                        "unit" : "x_unit",
                       "attributes" : {"unit" : {"storeas" : "content"}},
                      }
    SASDATA_IDATA_I = {"units_optional" : False,
                       "storeas" : "float",
                        "unit" : "y_unit",
                       "attributes" : {"unit" : {"storeas" : "content"}},
                      }
    SASDATA_IDATA_IDEV = {"units_optional" : False,
                          "storeas" : "float",
                          "unit" : "y_unit",
                          "attributes" : {"unit" : {"storeas" : "content"}},
                         }
    SASDATA_IDATA_QDEV = {"units_optional" : False,
                          "storeas" : "float",
                          "unit" : "x_unit",
                          "attributes" : {"unit" : {"storeas" : "content"}},
                         }
    SASDATA_IDATA_DQL = {"units_optional" : False,
                         "storeas" : "float",
                         "unit" : "x_unit",
                         "attributes" : {"unit" : {"storeas" : "content"}},
                        }
    SASDATA_IDATA_DQW = {"units_optional" : False,
                         "storeas" : "float",
                         "unit" : "x_unit",
                         "attributes" : {"unit" : {"storeas" : "content"}},
                        }
    SASDATA_IDATA_QMEAN = {"unit" : "x_unit",
                           "attributes" : {"unit" : {}},
                          }
    SASDATA_IDATA_SHADOWFACTOR = {}
    SASDATA_IDATA = {"attributes" : {"name" : {},"timestamp" : {"storeas" : "timestamp"}},
                     "children" : {"Q" : SASDATA_IDATA_Q,
                                   "I" : SASDATA_IDATA_I,
                                   "Idev" : SASDATA_IDATA_IDEV,
                                   "Qdev" : SASDATA_IDATA_QDEV,
                                   "dQw" : SASDATA_IDATA_DQW,
                                   "dQl" : SASDATA_IDATA_DQL,
                                   "Qmean" : SASDATA_IDATA_QMEAN,
                                   "Shadowfactor" : SASDATA_IDATA_SHADOWFACTOR,
                                   "<any>" : ANY
                                  }
                    }
    SASDATA = {"attributes" : {"name" : {}},
               "variable" : None,
               "children" : {"Idata" : SASDATA_IDATA,
                             "Sesans": {"storeas": "content"},
                             "zacceptance": {"storeas": "float"},
                             "yacceptance": {"storeas": "float"},
                             "<any>" : ANY
                            }
              }
    SASTRANSSPEC_TDATA_LAMDBA = {"storeas" : "float",
                                 "unit" : "wavelength_unit",
                                 "attributes" : {"unit" : {"storeas" : "content"}}
                                }
    SASTRANSSPEC_TDATA_T = {"storeas" : "float",
                            "unit" : "transmission_unit",
                            "attributes" : {"unit" : {"storeas" : "content"}}
                           }
    SASTRANSSPEC_TDATA_TDEV = {"storeas" : "float",
                               "unit" : "transmission_deviation_unit",
                               "attributes" : {"unit" :{"storeas" : "content"}}
                              }
    SASTRANSSPEC_TDATA = {"children" : {"Lambda" : SASTRANSSPEC_TDATA_LAMDBA,
                                        "T" : SASTRANSSPEC_TDATA_T,
                                        "Tdev" : SASTRANSSPEC_TDATA_TDEV,
                                        "<any>" : ANY,
                                       }
                         }
    SASTRANSSPEC = {"children" : {"Tdata" : SASTRANSSPEC_TDATA,
                                  "<any>" : ANY,
                                 },
                    "attributes" : {"name" :{}, "timestamp" : {},}
                   }
    SASSAMPLE_THICK = {"unit" : "thickness_unit",
                       "storeas" : "float",
                       "attributes" : {"unit" :{}},
                      }
    SASSAMPLE_TRANS = {"storeas" : "float",}
    SASSAMPLE_TEMP = {"unit" : "temperature_unit",
                      "storeas" : "float",
                      "attributes" :{"unit" :{}},
                     }
    SASSAMPLE_POS_ATTR = {"unit" : {}}
    SASSAMPLE_POS_X = {"unit" : "position_unit",
                       "storeas" : "float",
                       "attributes" : SASSAMPLE_POS_ATTR
                      }
    SASSAMPLE_POS_Y = {"unit" : "position_unit",
                       "storeas" : "float",
                       "attributes" : SASSAMPLE_POS_ATTR
                      }
    SASSAMPLE_POS_Z = {"unit" : "position_unit",
                       "storeas" : "float",
                       "attributes" : SASSAMPLE_POS_ATTR
                      }
    SASSAMPLE_POS = {"children" : {"x" : SASSAMPLE_POS_X,
                                   "y" : SASSAMPLE_POS_Y,
                                   "z" : SASSAMPLE_POS_Z,
                                  },
                    }
    SASSAMPLE_ORIENT_ATTR = {"unit" :{}}
    SASSAMPLE_ORIENT_ROLL = {"unit" : "orientation_unit",
                             "storeas" : "float",
                             "attributes" : SASSAMPLE_ORIENT_ATTR
                            }
    SASSAMPLE_ORIENT_PITCH = {"unit" : "orientation_unit",
                              "storeas" : "float",
                              "attributes" : SASSAMPLE_ORIENT_ATTR
                             }
    SASSAMPLE_ORIENT_YAW = {"unit" : "orientation_unit",
                            "storeas" : "float",
                            "attributes" : SASSAMPLE_ORIENT_ATTR
                           }
    SASSAMPLE_ORIENT = {"children" : {"roll" : SASSAMPLE_ORIENT_ROLL,
                                      "pitch" : SASSAMPLE_ORIENT_PITCH,
                                      "yaw" : SASSAMPLE_ORIENT_YAW,
                                     },
                       }
    SASSAMPLE = {"attributes" :
                 {"name" : {},},
                 "children" : {"ID" : {},
                               "thickness" : SASSAMPLE_THICK,
                               "transmission" : SASSAMPLE_TRANS,
                               "temperature" : SASSAMPLE_TEMP,
                               "position" : SASSAMPLE_POS,
                               "orientation" : SASSAMPLE_ORIENT,
                               "details" : {},
                               "<any>" : ANY
                              },
                }
    SASINSTR_SRC_BEAMSIZE_ATTR = {"unit" : ""}
    SASINSTR_SRC_BEAMSIZE_X = {"unit" : "beam_size_unit",
                               "storeas" : "float",
                               "attributes" : SASINSTR_SRC_BEAMSIZE_ATTR
                              }
    SASINSTR_SRC_BEAMSIZE_Y = {"unit" : "beam_size_unit",
                               "storeas" : "float",
                               "attributes" : SASINSTR_SRC_BEAMSIZE_ATTR
                              }
    SASINSTR_SRC_BEAMSIZE_Z = {"unit" : "beam_size_unit",
                               "storeas" : "float",
                               "attributes" : SASINSTR_SRC_BEAMSIZE_ATTR
                              }
    SASINSTR_SRC_BEAMSIZE = {"attributes" : {"name" : {}},
                             "children" : {"x" : SASINSTR_SRC_BEAMSIZE_X,
                                           "y" : SASINSTR_SRC_BEAMSIZE_Y,
                                           "z" : SASINSTR_SRC_BEAMSIZE_Z,
                                          }
                            }
    SASINSTR_SRC_WL = {"unit" : "wavelength_unit",
                       "storeas" : "float",
                       "attributes" : {"unit" :{},
                       }
                      }
    SASINSTR_SRC_WL_MIN = {"unit" : "wavelength_min_unit",
                           "storeas" : "float",
                           "attributes" : {"unit" :{"storeas" : "content"},}
                          }
    SASINSTR_SRC_WL_MAX = {"unit" : "wavelength_max_unit",
                           "storeas" : "float",
                           "attributes" : {"unit" :{"storeas" : "content"},}
                          }
    SASINSTR_SRC_WL_SPR = {"unit" : "wavelength_spread_unit",
                           "storeas" : "float",
                           "attributes" : {"unit" : {"storeas" : "content"},}
                          }
    SASINSTR_SRC = {"attributes" : {"name" : {}},
                    "children" : {"radiation" : {},
                                  "beam_size" : SASINSTR_SRC_BEAMSIZE,
                                  "beam_shape" : {},
                                  "wavelength" : SASINSTR_SRC_WL,
                                  "wavelength_min" : SASINSTR_SRC_WL_MIN,
                                  "wavelength_max" : SASINSTR_SRC_WL_MAX,
                                  "wavelength_spread" : SASINSTR_SRC_WL_SPR,
                                 },
                   }
    SASINSTR_COLL_APER_ATTR = {"unit" : {}}
    SASINSTR_COLL_APER_X = {"unit" : "size_unit",
                            "storeas" : "float",
                            "attributes" : SASINSTR_COLL_APER_ATTR
                           }
    SASINSTR_COLL_APER_Y = {"unit" : "size_unit",
                            "storeas" : "float",
                            "attributes" : SASINSTR_COLL_APER_ATTR
                           }
    SASINSTR_COLL_APER_Z = {"unit" : "size_unit",
                            "storeas" : "float",
                            "attributes" : SASINSTR_COLL_APER_ATTR
                           }
    SASINSTR_COLL_APER_SIZE = {"attributes" : {"unit" : {}},
                               "children" : {"storeas" : "float",
                                             "x" : SASINSTR_COLL_APER_X,
                                             "y" : SASINSTR_COLL_APER_Y,
                                             "z" : SASINSTR_COLL_APER_Z,
                                            }
                              }
    SASINSTR_COLL_APER_DIST = {"storeas" : "float",
                               "attributes" : {"unit" : {}},
                               "unit" : "distance_unit",
                              }
    SASINSTR_COLL_APER = {"attributes" : {"name" : {}, "type" : {}, },
                          "children" : {"size" : SASINSTR_COLL_APER_SIZE,
                                        "distance" : SASINSTR_COLL_APER_DIST
                                       }
                         }
    SASINSTR_COLL = {"attributes" : {"name" : {}},
                     "children" :
                         {"length" :
                          {"unit" : "length_unit",
                           "storeas" : "float",
                           "attributes" : {"storeas" : "content", "unit" : {}},
                          },
                          "aperture" : SASINSTR_COLL_APER,
                         },
                    }
    SASINSTR_DET_SDD = {"storeas" : "float",
                        "unit" : "distance_unit",
                        "attributes" : {"unit" :{}},
                       }
    SASINSTR_DET_OFF_ATTR = {"unit" : {"storeas" : "content" }}
    SASINSTR_DET_OFF_X = {"storeas" : "float",
                          "unit" : "offset_unit",
                          "attributes" : SASINSTR_DET_OFF_ATTR
                         }
    SASINSTR_DET_OFF_Y = {"storeas" : "float",
                          "unit" : "offset_unit",
                          "attributes" : SASINSTR_DET_OFF_ATTR
                         }
    SASINSTR_DET_OFF_Z = {"storeas" : "float",
                          "unit" : "offset_unit",
                          "attributes" : SASINSTR_DET_OFF_ATTR
                         }
    SASINSTR_DET_OFF = {"children" : {"x" : SASINSTR_DET_OFF_X,
                                      "y" : SASINSTR_DET_OFF_Y,
                                      "z" : SASINSTR_DET_OFF_Z,
                                     }
                       }
    SASINSTR_DET_OR_ATTR = {}
    SASINSTR_DET_OR_ROLL = {"storeas" : "float",
                            "unit" : "orientation_unit",
                            "attributes" : SASINSTR_DET_OR_ATTR
                           }
    SASINSTR_DET_OR_PITCH = {"storeas" : "float",
                             "unit" : "orientation_unit",
                             "attributes" : SASINSTR_DET_OR_ATTR
                            }
    SASINSTR_DET_OR_YAW = {"storeas" : "float",
                           "unit" : "orientation_unit",
                           "attributes" : SASINSTR_DET_OR_ATTR
                          }
    SASINSTR_DET_OR = {"children" : {"roll" : SASINSTR_DET_OR_ROLL,
                                     "pitch" : SASINSTR_DET_OR_PITCH,
                                     "yaw" : SASINSTR_DET_OR_YAW,
                                    }
                      }
    SASINSTR_DET_BC_X = {"storeas" : "float",
                         "unit" : "beam_center_unit",
                         "attributes" : {"storeas" : "content"}
                        }
    SASINSTR_DET_BC_Y = {"storeas" : "float",
                         "unit" : "beam_center_unit",
                         "attributes" : {"storeas" : "content"}
                        }
    SASINSTR_DET_BC_Z = {"storeas" : "float",
                         "unit" : "beam_center_unit",
                         "attributes" : {"storeas" : "content"}
                        }
    SASINSTR_DET_BC = {"children" : {"x" : SASINSTR_DET_BC_X,
                                     "y" : SASINSTR_DET_BC_Y,
                                     "z" : SASINSTR_DET_BC_Z,}
                      }
    SASINSTR_DET_PIXEL_X = {"storeas" : "float",
                            "unit" : "pixel_size_unit",
                            "attributes" : {"storeas" : "content" }
                           }
    SASINSTR_DET_PIXEL_Y = {"storeas" : "float",
                            "unit" : "pixel_size_unit",
                            "attributes" : {"storeas" : "content"}
                           }
    SASINSTR_DET_PIXEL_Z = {"storeas" : "float",
                            "unit" : "pixel_size_unit",
                            "attributes" : {"storeas" : "content"}
                           }
    SASINSTR_DET_PIXEL = {"children" : {"x" : SASINSTR_DET_PIXEL_X,
                                        "y" : SASINSTR_DET_PIXEL_Y,
                                        "z" : SASINSTR_DET_PIXEL_Z,
                                       }
                         }
    SASINSTR_DET_SLIT = {"storeas" : "float",
                         "unit" : "slit_length_unit",
                         "attributes" : {"unit" : {}}
                        }
    SASINSTR_DET = {"attributes" : {"name" : {"storeas" : "content"}},
                    "children" : {"name" : {"storeas" : "content"},
                                  "SDD" : SASINSTR_DET_SDD,
                                  "offset" : SASINSTR_DET_OFF,
                                  "orientation" : SASINSTR_DET_OR,
                                  "beam_center" : SASINSTR_DET_BC,
                                  "pixel_size" : SASINSTR_DET_PIXEL,
                                  "slit_length" : SASINSTR_DET_SLIT,
                                 }
                   }
    SASINSTR = {"children" :
                {"name" : {},
                 "SASsource" : SASINSTR_SRC,
                 "SAScollimation" : SASINSTR_COLL,
                 "SASdetector" : SASINSTR_DET,
                },
               }
    CANSAS_FORMAT = {"SASentry" :
                     {"units_optional" : True,
                      "storeas" : "content",
                      "attributes" : {"name" : {}},
                      "children" : {"Title" : TITLE,
                                    "Run" : RUN,
                                    "SASdata" : SASDATA,
                                    "SAStransmission_spectrum" : SASTRANSSPEC,
                                    "SASsample" : SASSAMPLE,
                                    "SASinstrument" : SASINSTR,
                                    "SASprocess" : SASPROCESS,
                                    "SASnote" : SASNOTE,
                                    "<any>" : ANY,
                                   }
                     }
                    }


class CurrentLevel:
    """
    A helper class to hold information on where you are in the constants tree
    """

    current_level = ''
    ns_datatype = ''
    ns_optional = True

    def __init__(self):
        self.current_level = {}
        self.ns_datatype = "content"
        self.ns_optional = True

    def get_current_level(self):
        """
        Helper method to get the current_level map
        """
        return self.current_level

    def get_data_type(self):
        """
        Helper method to get the ns_datatype label
        """
        return self.ns_datatype



"""
    Unit tests for data manipulations
"""
import unittest
from sasdata.dataloader.loader import Loader
from sasdata.dataloader.readers.abs_reader import Reader as AbsReader
from sasdata.dataloader.readers.danse_reader import Reader as DANSEReader

import os.path


def find(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)


class abs_reader(unittest.TestCase):

    def setUp(self):
        reader = AbsReader()
        self.data_list = reader.read(find("jan08002.ABS"))
        self.data = self.data_list[0]

    def test_abs_checkdata(self):
        """
            Check the data content to see whether
            it matches the specific file we loaded.
            Check the units too to see whether the
            Data1D defaults changed. Otherwise the
            tests won't pass
        """
        self.assertEqual(os.path.basename(self.data.filename), "jan08002.ABS")
        self.assertEqual(self.data.meta_data['loader'], "IGOR 1D")

        self.assertEqual(self.data.source.wavelength_unit, 'A')
        self.assertEqual(self.data.source.wavelength, 6.0)

        self.assertEqual(self.data.detector[0].distance_unit, 'mm')
        self.assertEqual(self.data.detector[0].distance, 1000.0)

        self.assertEqual(self.data.sample.transmission, 0.5667)

        self.assertEqual(self.data.detector[0].beam_center_unit, 'mm')
        center_x = 114.58*5.08
        center_y = 64.22*5.08
        self.assertEqual(self.data.detector[0].beam_center.x, center_x)
        self.assertEqual(self.data.detector[0].beam_center.y, center_y)

        self.assertEqual(self.data.y_unit, 'cm^{-1}')
        self.assertEqual(self.data.x[0], 0.008082)
        self.assertEqual(self.data.x[1], 0.0275)
        self.assertEqual(self.data.x[2], 0.02762)
        self.assertEqual(self.data.x[126], 0.5828)

        self.assertEqual(self.data.y[0], 0.02198)
        self.assertEqual(self.data.y[1], 0.02201)
        self.assertEqual(self.data.y[2], 0.02695)
        self.assertEqual(self.data.y[126], 0.2958)

        self.assertEqual(self.data.dy[0], 0.002704)
        self.assertEqual(self.data.dy[1], 0.001643)
        self.assertEqual(self.data.dy[2], 0.002452)
        self.assertEqual(self.data.dy[126], 1)

    def test_checkdata2(self):
        self.assertEqual(self.data.dy[126], 1)

    def test_generic_loader(self):
        # the generic loader should work as well
        data = Loader().load(find("jan08002.ABS"))
        self.assertEqual(data[0].meta_data['loader'], "IGOR 1D")

    def test_usans_negative_dxl(self):
        data_abs = Loader().load(find("sam14_cor.ABS"))[0]
        data_cor = Loader().load(find("sam14_cor.txt"))[0]
        for i in range(0, len(data_abs.x) - 1):
            self.assertEqual(data_abs.x[i], data_cor.x[i])
            self.assertEqual(data_abs.y[i], data_cor.y[i])
            self.assertEqual(data_abs.dxl[i], -data_cor.dx[i])
            self.assertTrue(data_abs.dxl[i] > 0)


class DanseReaderTests(unittest.TestCase):

    def setUp(self):
        reader = DANSEReader()
        self.data_list = reader.read(find("MP_New.sans"))
        self.data = self.data_list[0]

    def test_checkdata(self):
        """
            Check the data content to see whether
            it matches the specific file we loaded.
            Check the units too to see whether the
            Data1D defaults changed. Otherwise the
            tests won't pass
        """
        self.assertEqual(len(self.data_list), 1)
        self.assertEqual(os.path.basename(self.data.filename), "MP_New.sans")
        self.assertEqual(self.data.meta_data['loader'], "DANSE")

        self.assertEqual(self.data.source.wavelength_unit, 'A')
        self.assertEqual(self.data.source.wavelength, 7.5)

        self.assertEqual(self.data.detector[0].distance_unit, 'mm')
        self.assertAlmostEqual(self.data.detector[0].distance, 5414.99, 3)

        self.assertEqual(self.data.detector[0].beam_center_unit, 'mm')
        center_x = 68.74*5.0
        center_y = 64.77*5.0
        self.assertEqual(self.data.detector[0].beam_center.x, center_x)
        self.assertEqual(self.data.detector[0].beam_center.y, center_y)

        self.assertEqual(self.data.I_unit, 'cm^{-1}')
        self.assertEqual(self.data.data[0], 1.57831)
        self.assertEqual(self.data.data[1], 2.70983)
        self.assertEqual(self.data.data[2], 3.83422)

        self.assertEqual(self.data.err_data[0], 1.37607)
        self.assertEqual(self.data.err_data[1], 1.77569)
        self.assertEqual(self.data.err_data[2], 2.06313)

    def test_generic_loader(self):
        # the generic loader should work as well
        data = Loader().load(find("MP_New.sans"))
        self.assertEqual(len(data), 1)
        self.assertEqual(data[0].meta_data['loader'], "DANSE")


if __name__ == '__main__':
    unittest.main()



from typing import List, Any


def unique_preserve_order(seq: List[Any]) -> List[Any]:
    """ Remove duplicates from list preserving order
    Fastest according to benchmarks at https://www.peterbe.com/plog/uniqifiers-benchmark
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]



"""
    SESANS reader (based on ASCII reader)

    Reader for .ses or .sesans file format

    Jurrian Bakker
"""
import os

import numpy as np

from sasdata.dataloader.filereader import FileReader
from sasdata.dataloader.data_info import plottable_1D, DataInfo
from sasdata.data_util.loader_exceptions import FileContentsException

# Check whether we have a converter available
has_converter = True
try:
    from sasdata.data_util.nxsunit import Converter
except ImportError:
    has_converter = False


class Reader(FileReader):
    """
    Class to load sesans files (6 columns).
    """
    # File type
    type_name = "SESANS"

    # Wildcards
    type = ["SESANS files (*.ses)|*.ses",
            "SESANS files (*..sesans)|*.sesans"]
    # List of allowed extensions
    ext = ['.ses', '.SES', '.sesans', '.SESANS']

    # Flag to bypass extension check
    allow_all = True

    def get_file_contents(self):
        self.current_datainfo = DataInfo()
        self.current_dataset = plottable_1D(np.array([]), np.array([]))
        self.current_datainfo.isSesans = True
        self.output = []

        line = self.nextline()
        params = {}
        while line and not line.startswith("BEGIN_DATA"):
            terms = line.split()
            if len(terms) >= 2:
                params[terms[0]] = " ".join(terms[1:])
            line = self.nextline()
        self.params = params

        if "FileFormatVersion" not in self.params:
            raise FileContentsException("SES file missing FileFormatVersion")
        if float(self.params["FileFormatVersion"]) >= 2.0:
            raise FileContentsException("SASView only supports SES version 1")

        if "SpinEchoLength_unit" not in self.params:
            raise FileContentsException("SpinEchoLength has no units")
        if "Wavelength_unit" not in self.params:
            raise FileContentsException("Wavelength has no units")
        if params["SpinEchoLength_unit"] != params["Wavelength_unit"]:
            raise FileContentsException(
                "The spin echo data has rudely used "
                "different units for the spin echo length "
                "and the wavelength.  While sasview could "
                "handle this instance, it is a violation "
                "of the file format and will not be "
                "handled by other software.")

        headers = self.nextline().split()

        self._insist_header(headers, "SpinEchoLength")
        self._insist_header(headers, "Depolarisation")
        self._insist_header(headers, "Depolarisation_error")
        self._insist_header(headers, "Wavelength")

        data = np.loadtxt(self.f_open)

        if data.shape[1] != len(headers):
            raise FileContentsException(f"File has {len(headers)} headers, but {data.shape[1]} columns")

        if not data.size:
            raise FileContentsException("{} is empty".format(self.filepath))
        x = data[:, headers.index("SpinEchoLength")]
        if "SpinEchoLength_error" in headers:
            dx = data[:, headers.index("SpinEchoLength_error")]
        else:
            dx = x * 0.05
        lam = data[:, headers.index("Wavelength")]
        if "Wavelength_error" in headers:
            dlam = data[:, headers.index("Wavelength_error")]
        else:
            dlam = lam * 0.05
        y = data[:, headers.index("Depolarisation")]
        dy = data[:, headers.index("Depolarisation_error")]

        lam_unit = self._unit_fetch("Wavelength")
        x, x_unit = self._unit_conversion(x, "A",
                                          self._unit_fetch(
                                              "SpinEchoLength"))
        dx, dx_unit = self._unit_conversion(
            dx, lam_unit,
            self._unit_fetch("SpinEchoLength"))
        dlam, dlam_unit = self._unit_conversion(
            dlam, lam_unit,
            self._unit_fetch("Wavelength"))
        y_unit = self._unit_fetch("Depolarisation")

        self.current_dataset.x = x
        self.current_dataset.y = y
        self.current_dataset.lam = lam
        self.current_dataset.dy = dy
        self.current_dataset.dx = dx
        self.current_dataset.dlam = dlam
        self.current_datainfo.isSesans = True

        self.current_datainfo._yunit = y_unit
        self.current_datainfo._xunit = x_unit
        self.current_datainfo.source.wavelength_unit = lam_unit
        self.current_datainfo.source.wavelength = lam
        self.current_datainfo.filename = os.path.basename(self.f_open.name)
        self.current_dataset.xaxis(r"\rm{z}", x_unit)
        # Adjust label to ln P/(lam^2 t), remove lam column refs
        self.current_dataset.yaxis(r"\rm{ln(P)/(t \lambda^2)}", y_unit)
        # Store loading process information
        self.current_datainfo.meta_data['loader'] = self.type_name
        self.current_datainfo.sample.name = params["Sample"]
        self.current_datainfo.sample.ID = params["DataFileTitle"]
        self.current_datainfo.sample.thickness = self._unit_conversion(
            float(params["Thickness"]), "cm",
            self._unit_fetch("Thickness"))[0]

        self.current_datainfo.sample.zacceptance = (
            float(params["Theta_zmax"]),
            self._unit_fetch("Theta_zmax"))

        self.current_datainfo.sample.yacceptance = (
            float(params["Theta_ymax"]),
            self._unit_fetch("Theta_ymax"))

        self.send_to_output()

    @staticmethod
    def _insist_header(headers: list, name: str):
        if name not in headers:
            raise FileContentsException(f"Missing {name} column in spin echo data")

    @staticmethod
    def _unit_conversion(value: float, value_unit: str, default_unit: str) -> (float, str):
        """
        Performs unit conversion on a measurement.

        :param value: The magnitude of the measurement
        :param value_unit: a string containing the final desired unit
        :param default_unit: string with the units of the original measurement
        :return: The magnitude of the measurement in the new units
        """
        # (float, string, string) -> float
        if has_converter and value_unit != default_unit:
            data_conv_q = Converter(default_unit)
            value = data_conv_q(value, units=value_unit)
            new_unit = default_unit
        else:
            new_unit = value_unit
        return value, new_unit

    def _unit_fetch(self, unit: str) -> str:
        return self.params[unit+"_unit"]



"""
ASCII 2D Loader
"""
from sasdata.dataloader.data_info import Data2D
import numpy as np

# ISIS 2D ASCII File Format
# http://www.isis.stfc.ac.uk/instruments/loq/software/colette-ascii-file-format-descriptions9808.pdf
# line: property
# 0: File header
# 1: q_x axis label and units
# 2: q_y axis label and units
# 3: Intensity axis label and units
# 4: n_use_rec - number of lines of user content following this line
# 5 to (5+n_use_rec): user content
# Number of qx points
# List of qx points
# Number of qy points
# List of qy points
# numqx numqy scale

class ASCII2DLoader:

    def __init__(self, data_path):
        """
        :param data_path: The path to the file to load
        """
        self.data_path = data_path

    def load(self):
        """
        Load the data from the file into a Data2D object

        :return: A Data2D instance containing data from the file
        :raises ValueError: Raises a ValueError if the file is incorrectly
            formatted
        """
        with open(self.data_path, 'r') as file_handle:
            file_buffer = file_handle.read()
            all_lines = file_buffer.splitlines()

            # Load num_points line-by-line from lines into a numpy array,
            # starting on line number start_line
            def _load_points(lines, start_line, num_points):
                qs = np.zeros(num_points)
                n = start_line
                filled = 0
                while filled < num_points:
                    row = np.fromstring(lines[n], dtype=np.float32, sep=' ')
                    qs[filled:filled+len(row)] = row
                    filled += len(row)
                    n += 1
                return n, qs

            current_line = 4
            try:
                # Skip n_use_rec lines
                n_use_rec = int(all_lines[current_line].strip()[0])
                current_line += n_use_rec + 1
                # Read qx data
                num_qs = int(all_lines[current_line].strip())
                current_line += 1
                current_line, qx = _load_points(all_lines, current_line, num_qs)

                # Read qy data
                num_qs = int(all_lines[current_line].strip())
                current_line += 1
                current_line, qy = _load_points(all_lines, current_line, num_qs)
            except ValueError as e:
                err_msg = "File incorrectly formatted.\n"
                if str(e).find('broadcast') != -1:
                    err_msg += "Incorrect number of q data points provided. "
                    err_msg += "Expected {}.".format(num_qs)
                elif str(e).find('invalid literal') != -1:
                    err_msg += ("Expected integer on line {}. "
                        "Instead got '{}'").format(current_line + 1,
                            all_lines[current_line])
                else:
                    err_msg += str(e)
                raise ValueError(err_msg)

            # dimensions: [width, height, scale]
            try:
                dimensions = np.fromstring(all_lines[current_line],
                    dtype=np.float32, sep=' ')
                if len(dimensions) != 3: raise ValueError()
                width = int(dimensions[0])
                height = int(dimensions[1])
            except ValueError as e:
                err_msg = "File incorrectly formatted.\n"
                err_msg += ("Expected line {} to be of the form: <num_qx> "
                    "<num_qy> <scale>.").format(current_line + 1)
                err_msg += " Instead got '{}'.".format(all_lines[current_line])
                raise ValueError(err_msg)

            if width > len(qx) or height > len(qy):
                err_msg = "File incorrectly formatted.\n"
                err_msg += ("Line {} says to use {}x{} points. "
                    "Only {}x{} provided.").format(current_line + 1, width,
                    height, len(qx), len(qy))
                raise ValueError(err_msg)

            # More qx and/or qy points can be provided than are actually used
            qx = qx[:width]
            qy = qy[:height]

            current_line += 1
            # iflag = 1 => Only intensity data (not dealt with here)
            # iflag = 2 => q axis and intensity data
            # iflag = 3 => q axis, intensity and error data
            try:
                iflag = int(all_lines[current_line].strip()[0])
                if iflag <= 0 or iflag > 3: raise ValueError()
            except:
                err_msg = "File incorrectly formatted.\n"
                iflag = all_lines[current_line].strip()[0]
                err_msg += ("Expected iflag on line {} to be 1, 2 or 3. "
                    "Instead got '{}'.").format(current_line+1, iflag)
                raise ValueError(err_msg)

            current_line += 1

            try:
                current_line, I = _load_points(all_lines, current_line,
                    width * height)
                dI = np.zeros(width*height)

                # Load error data if it's provided
                if iflag == 3:
                    _, dI = _load_points(all_lines, current_line, width*height)
            except Exception as e:
                err_msg = "File incorrectly formatted.\n"
                if str(e).find("list index") != -1:
                    err_msg += ("Incorrect number of data points. Expected {}"
                        " intensity").format(width * height)
                    if iflag == 3:
                        err_msg += " and error"
                    err_msg += " points."
                else:
                    err_msg += str(e)
                raise ValueError(err_msg)

            # Format data for use with Data2D
            qx = list(qx) * height
            qy = np.array([[y] * width for y in qy]).flatten()

            data = Data2D(qx_data=qx, qy_data=qy, data=I, err_data=dI)

        return data



"""
    Unit tests for the ascii (n-column) reader
"""

import os.path
import warnings
import math

import unittest
from sasdata.dataloader.loader import Loader
from sasdata.dataloader.data_info import Data2D

warnings.simplefilter("ignore")


def find(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)


class ABSReaderTests(unittest.TestCase):
    
    def setUp(self):
        self.loader = Loader()
        self.f1_list = self.loader.load(find("ascii_test_1.txt"))
        self.f1 = self.f1_list[0]
        self.f2_list = self.loader.load(find("ascii_test_2.txt"))
        self.f2 = self.f2_list[0]
        self.f3_list = self.loader.load(find("ascii_test_3.txt"))
        self.f3 = self.f3_list[0]
        self.f4_list = self.loader.load(find("ascii_test_4.abs"))
        self.f4 = self.f4_list[0]
        self.f5_list = self.loader.load(find("ascii_test_5.txt"))
        self.f5 = self.f5_list[0]

    def test_checkdata(self):
        """
            Test .ABS file loaded as ascii
        """
        # The length of the data is 10
        self.assertEqual(len(self.f1_list), 1)
        self.assertEqual(len(self.f2_list), 1)
        self.assertEqual(len(self.f3_list), 1)
        self.assertEqual(len(self.f4_list), 1)
        self.assertEqual(len(self.f5_list), 1)
        self.assertEqual(len(self.f1.x), 10)
        self.assertEqual(self.f1.x[0],0.002618)
        self.assertEqual(self.f1.x[9],0.0497)
        self.assertTrue(self.f1.x_unit == 'A^{-1}')
        self.assertTrue(self.f1.y_unit == 'cm^{-1}')
        
        self.assertEqual(self.f1.meta_data['loader'],"ASCII")

    def test_truncated_1(self):
        """
            Test an ascii file with header and a 
            comment line in the middle of the data section.
            The business rule says that we should stop
            reading at the first comment once the data
            section has started (and treat the comment
            as though it were the start of a footer).
        """
        # The length of the data is 5
        self.assertEqual(len(self.f2.x), 5)
        self.assertEqual(self.f2.x[0],0.002618)
        self.assertEqual(self.f2.x[4],0.02356)

    def test_truncated_2(self):
        """
            Test a 6-col ascii file with header and a 
            line with only 2 columns in the middle of the data section.
            The business rule says that we should stop
            reading at the first inconsitent line.
        """
        # The length of the data is 5
        self.assertEqual(len(self.f3.x), 5)
        self.assertEqual(self.f3.x[0],0.002618)
        self.assertEqual(self.f3.x[4],0.02356)

    def test_truncated_3(self):
        """
            Test a 6-col ascii file with complex header and 
            many lines with 2 or 2 columns in the middle of the data section.
            The business rule says that we should stop
            reading at the last line of header.
        """
        # The length of the data is 5
        self.assertEqual(len(self.f4.x), 5)
        self.assertEqual(self.f4.x[0],0.012654)
        self.assertEqual(self.f4.x[4],0.02654)

    def test_truncated_4(self):
        """
            Test mix of 6-col and 2-col.
            Only the last 5 2-col lines should be read.
        """
        # The length of the data is 5
        self.assertEqual(len(self.f5.x), 5)
        self.assertEqual(self.f5.x[0],0.02879)
        self.assertEqual(self.f5.x[4],0.0497)

    def test_truncated_5(self):
        """
            Test a 6-col ascii file with complex header where one of them has a
            letter and many lines with 2 or 2 columns in the middle of the data
            section. Will be rejected because fewer than 5 lines.
        """
        # Test .ABS file loaded as ascii
        f = None
        try:
            f = self.loader.load(find("ascii_test_6.txt"))
        # The length of the data is 5
        except:
            self.assertEqual(f, None)

    def test_nan_values(self):
        """
        Test loading an ascii data file with nan values saved in x, y, or dy.
        """
        f_1d = self.loader.load(find("nans_in_1d_data.dat"))[0]
        f_2d = self.loader.load(find("nans_in_2d_data.DAT"))[0]
        for i in range(0, len(f_1d.x) - 1):
            self.assertFalse(math.isnan(f_1d.x[i]))
            self.assertFalse(math.isnan(f_1d.y[i]))
            self.assertFalse(math.isnan(f_1d.dy[i]))
        self.assertTrue(isinstance(f_2d, Data2D))
        f_2d.data = f_2d.data.flatten()
        f_2d.qx_data = f_2d.qx_data.flatten()
        f_2d.qy_data = f_2d.qy_data.flatten()
        for i in range(0, len(f_2d.data) - 1):
            self.assertFalse(math.isnan(f_2d.data[i]))
            self.assertFalse(math.isnan(f_2d.qx_data[i]))
            self.assertFalse(math.isnan(f_2d.qy_data[i]))

    def test_encoding(self):
        # Compare loading of utf-8 versus ansi/windows-1252
        ansi = self.loader.load(find("encoding_ANSI.txt"))
        utf8 = self.loader.load(find("encoding_UTF_8.txt"))
        self._check_cl_data(ansi)
        self._check_cl_data(utf8)

    def _check_cl_data(self, data):
        self.assertEqual(len(data), 1)
        data_set = data[0]
        self.assertEqual(len(data_set.x), 969)
        self.assertEqual(len(data_set.x), len(data_set.y))
        self.assertEqual(data_set.x[0], 0.000201634)
        self.assertEqual(data_set.y[0], 208487062.3)
        self.assertEqual(data_set.x_unit, 'A^{-1}')
        self.assertEqual(data_set.y_unit, 'cm^{-1}')

    def test_save_ascii(self):
        f_name = "test_f1_output.txt"
        f_name_csv = "test_f1_output.csv"
        # Save and load text file
        self.loader.save(f_name, self.f1, None)
        self.assertTrue(os.path.isfile(f_name))
        reload = self.loader.load(f_name)
        f1_reload = reload[0]
        # Compare data from saved text file to previously loaded data
        self.assertEqual(len(self.f1.x), len(f1_reload.x))
        self.assertEqual(len(self.f1.y), len(f1_reload.y))
        self.assertEqual(self.f1.y[0], f1_reload.y[0])
        # Save and load csv file
        self.loader.save(f_name_csv, self.f1, None)
        self.assertTrue(os.path.isfile(f_name_csv))
        reload_csv = self.loader.load(f_name_csv)
        f1_reload_csv = reload_csv[0]
        # Compare data from saved csv file to previously loaded data
        self.assertEqual(len(self.f1.x), len(f1_reload_csv.x))
        self.assertEqual(len(self.f1.y), len(f1_reload_csv.y))
        self.assertEqual(self.f1.y[0], f1_reload_csv.y[0])
        # Destroy generated files
        os.remove(f_name)
        self.assertFalse(os.path.isfile(f_name))
        os.remove(f_name_csv)
        self.assertFalse(os.path.isfile(f_name_csv))


if __name__ == '__main__':
    unittest.main()
   



"""
    Unit tests for data manipulations
"""

import math
import os.path
import unittest

import numpy as np
from numpy.testing import assert_allclose

from sasdata.dataloader.loader import Loader
from sasdata.dataloader.data_info import Data1D, Data2D

RTOL = 1e-12


def find(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)


class DataInfoTests(unittest.TestCase):

    def setUp(self):
        data = Loader().load(find("cansas1d.xml"))
        self.data = data[0]

    def test_clone1D(self):
        """
            Test basic cloning
        """
        clone = self.data.clone_without_data()

        for i in range(len(self.data.detector)):
            self.assertEqual(
                self.data.detector[i].distance, clone.detector[i].distance)


class Theory1DTests(unittest.TestCase):

    def setUp(self):
        data = Loader().load(find("cansas1d.xml"))
        self.data = data[0]

    def test_clone_theory1D(self):
        """
            Test basic cloning
        """
        theory = Data1D(x=[], y=[], dy=None)
        theory.clone_without_data(clone=self.data)
        theory.copy_from_datainfo(data1d=self.data)
        for i in range(len(self.data.detector)):
            self.assertEqual(
                self.data.detector[i].distance, theory.detector[i].distance)

        for i in range(len(self.data.x)):
            self.assertEqual(self.data.x[i], theory.x[i])
            self.assertEqual(self.data.y[i], theory.y[i])
            self.assertEqual(self.data.dy[i], theory.dy[i])


class ManipTests(unittest.TestCase):

    def setUp(self):
        # Create two data sets to play with
        x_0 = np.arange(1, 6, dtype='d')
        y_0 = 2.0 * np.ones(5)
        dy_0 = 0.5 * np.ones(5)
        self.data = Data1D(x_0, y_0, dy=dy_0)

        x = self.data.x.copy()
        y = np.ones(5)
        dy = np.ones(5)
        self.data2 = Data1D(x, y, dy=dy)

    def test_load(self):
        """
            Test whether the test file was loaded properly
        """
        # There should be 5 entries in the file
        self.assertEqual(len(self.data.x), 5)

        for i in range(5):
            # The x values should be from 1 to 5
            self.assertEqual(self.data.x[i], float(i + 1))

        # All y-error values should be 0.5
        assert_allclose(self.data.dy, 0.5, RTOL)

        # All y values should be 2.0
        assert_allclose(self.data.y, 2.0, RTOL)

    def test_add(self):
        result = self.data2 + self.data
        assert_allclose(result.y, 3.0, RTOL)
        assert_allclose(result.dy, np.sqrt(0.5**2 + 1.0), RTOL)

    def test_sub(self):
        result = self.data2 - self.data
        assert_allclose(result.y, -1.0, RTOL)
        assert_allclose(result.dy, math.sqrt(0.5**2 + 1.0), RTOL)

    def test_mul(self):
        result = self.data2 * self.data
        assert_allclose(result.y, 2.0, RTOL)
        assert_allclose(
            result.dy, math.sqrt((0.5 * 1.0)**2 + (1.0 * 2.0)**2), RTOL)

    def test_div(self):
        result = self.data2 / self.data
        assert_allclose(result.y, 0.5, RTOL)
        assert_allclose(
            result.dy, math.sqrt((1.0 / 2.0)**2 + (0.5 * 1.0 / 4.0)**2), RTOL)

    def test_radd(self):
        result = self.data + 3.0
        assert_allclose(result.y, 5.0, RTOL)
        assert_allclose(result.dy, 0.5, RTOL)

        result = 3.0 + self.data
        assert_allclose(result.y, 5.0, RTOL)
        assert_allclose(result.dy, 0.5, RTOL)

    def test_rsub(self):
        result = self.data - 3.0
        assert_allclose(result.y, -1.0, RTOL)
        assert_allclose(result.dy, 0.5, RTOL)

        result = 3.0 - self.data
        assert_allclose(result.y, 1.0, RTOL)
        assert_allclose(result.dy, 0.5, RTOL)

    def test_rmul(self):
        result = self.data * 3.0
        assert_allclose(result.y, 6.0, RTOL)
        assert_allclose(result.dy, 1.5, RTOL)

        result = 3.0 * self.data
        assert_allclose(result.y, 6.0, RTOL)
        assert_allclose(result.dy, 1.5, RTOL)

    def test_rdiv(self):
        result = self.data / 4.0
        assert_allclose(result.y, 0.5, RTOL)
        assert_allclose(result.dy, 0.125, RTOL)

        result = 6.0 / self.data
        assert_allclose(result.y, 3.0, RTOL)
        assert_allclose(result.dy, 6.0 * 0.5 / 4.0, RTOL)


class Manin2DTests(unittest.TestCase):

    def setUp(self):
        # Create two data sets to play with
        x_0 = 2.0 * np.ones(25)
        dx_0 = 0.5 * np.ones(25)
        qx_0 = np.arange(25)
        qy_0 = np.arange(25)
        mask_0 = np.zeros(25)
        dqx_0 = np.arange(25) / 100
        dqy_0 = np.arange(25) / 100
        q_0 = np.sqrt(qx_0 * qx_0 + qy_0 * qy_0)
        self.data = Data2D(data=x_0, err_data=dx_0, qx_data=qx_0,
                           qy_data=qy_0, q_data=q_0, mask=mask_0,
                           dqx_data=dqx_0, dqy_data=dqy_0)

        y = np.ones(25)
        dy = np.ones(25)
        qx = np.arange(25)
        qy = np.arange(25)
        mask = np.zeros(25)
        q = np.sqrt(qx * qx + qy * qy)
        self.data2 = Data2D(data=y, err_data=dy, qx_data=qx, qy_data=qy,
                            q_data=q, mask=mask)

    def test_load(self):
        """
            Test whether the test file was loaded properly
        """
        # There should be 5 entries in the file
        self.assertEqual(np.size(self.data.data), 25)

        for i in range(25):
            # All y-error values should be 0.5
            self.assertEqual(self.data.err_data[i], 0.5)

            # All y values should be 2.0
            self.assertEqual(self.data.data[i], 2.0)

    def test_add(self):
        result = self.data2 + self.data
        assert_allclose(result.data, 3.0, RTOL)
        assert_allclose(result.err_data, np.sqrt(0.5**2 + 1.0), RTOL)

    def test_sub(self):
        result = self.data2 - self.data
        assert_allclose(result.data, -1.0, RTOL)
        assert_allclose(result.err_data, math.sqrt(0.5**2 + 1.0), RTOL)

    def test_mul(self):
        result = self.data2 * self.data
        assert_allclose(result.data, 2.0, RTOL)
        assert_allclose(
            result.err_data, math.sqrt((0.5 * 1.0)**2 + (1.0 * 2.0)**2), RTOL)

    def test_div(self):
        result = self.data2 / self.data
        assert_allclose(result.data, 0.5, RTOL)
        assert_allclose(
            result.err_data,
            math.sqrt((1.0 / 2.0)**2 + (0.5 * 1.0 / 4.0)**2), RTOL)

    def test_radd(self):
        result = self.data + 3.0
        assert_allclose(result.data, 5.0, RTOL)
        assert_allclose(result.err_data, 0.5, RTOL)

        result = 3.0 + self.data
        assert_allclose(result.data, 5.0, RTOL)
        assert_allclose(result.err_data, 0.5, RTOL)

    def test_rsub(self):
        result = self.data - 3.0
        assert_allclose(result.data, -1.0, RTOL)
        assert_allclose(result.err_data, 0.5, RTOL)

        result = 3.0 - self.data
        assert_allclose(result.data, 1.0, RTOL)
        assert_allclose(result.err_data, 0.5, RTOL)

    def test_rmul(self):
        result = self.data * 3.0
        assert_allclose(result.data, 6.0, RTOL)
        assert_allclose(result.err_data, 1.5, RTOL)

        result = 3.0 * self.data
        assert_allclose(result.data, 6.0, RTOL)
        assert_allclose(result.err_data, 1.5, RTOL)

    def test_rdiv(self):
        result = self.data / 4.0
        assert_allclose(result.data, 0.5, RTOL)
        assert_allclose(result.err_data, 0.125, RTOL)

        result = 6.0 / self.data
        assert_allclose(result.data, 3.0, RTOL)
        assert_allclose(result.err_data, 6.0 * 0.5 / 4.0, RTOL)


if __name__ == '__main__':
    unittest.main()



from sasdata.file_converter.nxcansas_writer import NXcanSASWriter
from sasdata.dataloader.loader import Loader

import os
import os.path
import unittest
import warnings

warnings.simplefilter("ignore")


def find(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)


class nxcansas_writer(unittest.TestCase):

    def setUp(self):
        self.loader = Loader()
        self.writer = NXcanSASWriter()
        self.read_file_1d = find("cansas1d.xml")
        self.write_file_1d = find("export1d.h5")
        self.read_file_2d = find("exp18_14_igor_2dqxqy.dat")
        self.write_file_2d = find("export2d.h5")

        self.data_1d = self.loader.load(self.read_file_1d)[0]

        self.data_2d = self.loader.load(self.read_file_2d)[0]
        self.data_2d.detector[0].name = ''
        self.data_2d.source.radiation = 'neutron'

    def test_write_1d(self):
        self.writer.write([self.data_1d], self.write_file_1d)
        data = self.loader.load(self.write_file_1d)
        self.assertTrue(len(data) == 1)
        data = data[0]
        self.assertTrue(len(data.x) == len(self.data_1d.x))
        self.assertTrue(len(data.y) == len(self.data_1d.y))
        self.assertTrue(len(data.dy) == len(self.data_1d.dy))
        self._check_metadata(data, self.data_1d)

    def test_write_2d(self):
        self.writer.write([self.data_2d], self.write_file_2d)
        data = self.loader.load(self.write_file_2d)
        self.assertTrue(len(data) == 1)
        data = data[0]
        self.assertTrue(len(data.data) == len(self.data_2d.data))
        self.assertTrue(len(data.qx_data) == len(self.data_2d.qx_data))
        self.assertTrue(len(data.qy_data) == len(self.data_2d.qy_data))
        self._check_metadata(data, self.data_2d)

    def _check_metadata(self, written, correct):
        self.assertTrue(written.title == correct.title)
        self.assertTrue(written.sample.name == correct.sample.name)
        self.assertAlmostEqual(written.sample.thickness, correct.sample.thickness)
        self.assertAlmostEqual(written.sample.temperature, correct.sample.temperature)
        self.assertTrue(written.instrument == correct.instrument)
        self.assertTrue(len(written.detector) == len(correct.detector))
        for i in range(len(written.detector)):
            written_det = written.detector[i]
            correct_det = correct.detector[i]
            self.assertAlmostEqual(written_det.distance, correct_det.distance)
            self.assertTrue(written_det.name == correct_det.name)
        self.assertTrue(written.source.probe == correct.source.radiation)

    def tearDown(self):
        if os.path.isfile(self.write_file_1d):
            os.remove(self.write_file_1d)
        if os.path.isfile(self.write_file_2d):
            os.remove(self.write_file_2d)



import os
import os.path
import unittest

from xml.etree import ElementTree as ET
import numpy as np

from sasdata.file_converter.bsl_loader import BSLLoader


def find(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)


class bsl_test(unittest.TestCase):

    def setUp(self):
        self.read_file_q = find("Z83000.QAX")
        self.read_file_i = find("Z83000.I1D")

        self.q_reader = BSLLoader(self.read_file_q)
        self.i_reader = BSLLoader(self.read_file_i)

    def test_load(self):
        #Load the data from bsl files.
        q_data_load = self.q_reader.load_data(0)
        i_data_load = self.i_reader.load_data(0)

        #Load the correct data (in Z83000.xml)

        tree = ET.parse(find('Z83000.xml'))
        root = tree.getroot()

        q_data_correct = [elem.text for elem in root.iter('{cansas1d/1.0}Q')]
        i_data_correct = [elem.text for elem in root.iter('{cansas1d/1.0}I')]

        q_data_array = np.array(q_data_correct, dtype=np.float64)
        i_data_array = np.array(i_data_correct, dtype=np.float64)

        q_test = np.allclose(q_data_array, q_data_load, atol=1e-13)
        q_test = np.allclose(i_data_array, i_data_load, atol=1e-13)

        self.assertTrue(q_test)



"""
Exceptions specific to loading data.
"""

from typing import Optional


class NoKnownLoaderException(Exception):
    """
    Exception for files with no associated reader based on the file
    extension of the loaded file. This exception should only be thrown by
    loader.py.
    """
    def __init__(self, e: Optional[str] = None):
        self.message = e


class DefaultReaderException(Exception):
    """
    Exception for files with no associated reader. This should be thrown by
    default readers only to tell Loader to try the next reader.
    """
    def __init__(self, e: Optional[str] = None):
        self.message = e


class FileContentsException(Exception):
    """
    Exception for files with an associated reader, but with no loadable data.
    This is useful for catching loader or file format issues.
    """
    def __init__(self, e: Optional[str] = None):
        self.message = e


class DataReaderException(Exception):
    """
    Exception for files that were able to mostly load, but had minor issues
    along the way.
    Any exceptions of this type should be put into the datainfo.errors
    """
    def __init__(self, e: Optional[str] = None):
        self.message = e



"""
    NXcanSAS 1/2D data reader for writing HDF5 formatted NXcanSAS files.
"""

import h5py
import numpy as np

from sasdata.dataloader.data_info import Data1D, Data2D

PROBES = ['neutron', 'x-ray', 'muon', 'electron', 'ultraviolet',
          'visible light', 'positron', 'proton']
TYPES = ['Spallation Neutron Source', 'UV Plasma Source', 'Free-Electron Laser',
         'Reactor Neutron Source', 'Synchrotron X-ray Source', 'UV Laser',
         'Pulsed Muon Source', 'Pulsed Reactor Neutron Source', 'Ion Source',
         'Rotating Anode X-ray', 'Optical Laser', 'Fixed Tube X-ray']


class NXcanSASWriter:
    """
    A class for writing in NXcanSAS data files. Any number of data sets may be
    written to the file. Currently 1D and 2D SAS data sets are supported

    NXcanSAS spec: http://download.nexusformat.org/sphinx/classes/contributed_definitions/NXcanSAS.html

    :Dependencies:
        The NXcanSAS writer requires h5py => v2.5.0 or later.
    """

    def write(self, dataset, filename):
        """
        Write an array of Data1d or Data2D objects to an NXcanSAS file, as
        one SASEntry with multiple SASData elements. The metadata of the first
        elememt in the array will be written as the SASentry metadata
        (detector, instrument, sample, etc).

        :param dataset: A list of Data1D or Data2D objects to write
        :param filename: Where to write the NXcanSAS file
        """

        def _h5_string(string):
            """
            Convert a string to a numpy string in a numpy array. This way it is
            written to the HDF5 file as a fixed length ASCII string and is
            compatible with the Reader read() method.
            """
            if isinstance(string, np.ndarray):
                return string
            elif not isinstance(string, str):
                string = str(string)

            return np.array([np.string_(string)])

        def _write_h5_string(entry, value, key):
            entry[key] = _h5_string(value)

        def _h5_float(x):
            if not (isinstance(x, list)):
                x = [x]
            return np.array(x, dtype=np.float32)

        def _write_h5_float(entry, value, key):
            entry.create_dataset(key, data=_h5_float(value))

        def _write_h5_vector(entry, vector, names=['x_position', 'y_position'],
            units=None, write_fn=_write_h5_string):
            """
            Write a vector to an h5 entry

            :param entry: The H5Py entry to write to
            :param vector: The Vector to write
            :param names: What to call the x,y and z components of the vector
                when writing to the H5Py entry
            :param units: The units of the vector (optional)
            :param write_fn: A function to convert the value to the required
                format and write it to the H5Py entry, of the form
                f(entry, value, name) (optional)
            """
            if len(names) < 2:
                raise ValueError("Length of names must be >= 2.")

            if vector.x is not None:
                write_fn(entry, vector.x, names[0])
                if units is not None:
                    entry[names[0]].attrs['units'] = units
            if vector.y is not None:
                write_fn(entry, vector.y, names[1])
                if units is not None:
                    entry[names[1]].attrs['units'] = units
            if len(names) == 3 and vector.z is not None:
                write_fn(entry, vector.z, names[2])
                if units is not None:
                    entry[names[2]].attrs['units'] = units

        valid_data = all([isinstance(d, (Data1D, Data2D)) for d in dataset])
        if not valid_data:
            raise ValueError("All entries of dataset must be Data1D or Data2D"
                             "objects")

        # Get run name and number from first Data object
        data_info = dataset[0]
        run_number = ''
        run_name = ''
        if len(data_info.run) > 0:
            run_number = data_info.run[0]
            if len(data_info.run_name) > 0:
                run_name = data_info.run_name[run_number]

        f = h5py.File(filename, 'w')
        sasentry = f.create_group('sasentry01')
        sasentry['definition'] = _h5_string('NXcanSAS')
        sasentry['run'] = _h5_string(run_number)
        sasentry['run'].attrs['name'] = run_name
        sasentry['title'] = _h5_string(data_info.title)
        sasentry.attrs['canSAS_class'] = 'SASentry'
        sasentry.attrs['version'] = '1.1'

        for i, data_obj in enumerate(dataset):
            data_entry = sasentry.create_group("sasdata{0:0=2d}".format(i+1))
            data_entry.attrs['canSAS_class'] = 'SASdata'
            if isinstance(data_obj, Data1D):
                self._write_1d_data(data_obj, data_entry)
            elif isinstance(data_obj, Data2D):
                self._write_2d_data(data_obj, data_entry)

        data_info = dataset[0]
        # Sample metadata
        sample_entry = sasentry.create_group('sassample')
        sample_entry.attrs['canSAS_class'] = 'SASsample'
        sample_entry['ID'] = _h5_string(data_info.sample.name)
        sample_attrs = ['thickness', 'temperature', 'transmission']
        for key in sample_attrs:
            if getattr(data_info.sample, key) is not None:
                sample_entry.create_dataset(key,
                    data=_h5_float(getattr(data_info.sample, key)))
        _write_h5_vector(sample_entry, data_info.sample.position)
        # NXcanSAS doesn't save information about pitch, only roll
        # and yaw. The _write_h5_vector method writes vector.y, but we
        # need to write vector.z for yaw
        data_info.sample.orientation.y = data_info.sample.orientation.z
        _write_h5_vector(sample_entry, data_info.sample.orientation,
            names=['polar_angle', 'azimuthal_angle'])
        if data_info.sample.details is not None\
            and data_info.sample.details != []:
            details = None
            if len(data_info.sample.details) > 1:
                details = [np.string_(d) for d in data_info.sample.details]
                details = np.array(details)
            elif data_info.sample.details != []:
                details = _h5_string(data_info.sample.details[0])
            if details is not None:
                sample_entry.create_dataset('details', data=details)

        # Instrument metadata
        instrument_entry = sasentry.create_group('sasinstrument')
        instrument_entry.attrs['canSAS_class'] = 'SASinstrument'
        instrument_entry['name'] = _h5_string(data_info.instrument)

        # Source metadata
        source_entry = instrument_entry.create_group('sassource')
        source_entry.attrs['canSAS_class'] = 'SASsource'
        if data_info.source.radiation is not None:
            if data_info.source.radiation in PROBES:
                source_entry['probe'] = _h5_string(data_info.source.radiation)
            elif data_info.source.radiation in TYPES:
                source_entry['type'] = _h5_string(data_info.source.radiation)
            else:
                # Should not get here, but we shouldn't throw info out either
                source_entry['notes'] = _h5_string(data_info.source.radiation)
        if data_info.source.type is not None:
            source_entry['type'] = _h5_string(data_info.source.type)
        if data_info.source.probe is not None:
            source_entry['probe'] = _h5_string(data_info.source.probe)
        if data_info.source.beam_shape is not None:
            source_entry['beam_shape'] = _h5_string(data_info.source.beam_shape)
        wavelength_keys = { 'wavelength': 'incident_wavelength',
            'wavelength_min':'wavelength_min',
            'wavelength_max': 'wavelength_max',
            'wavelength_spread': 'incident_wavelength_spread' }
        for sasname, nxname in wavelength_keys.items():
            value = getattr(data_info.source, sasname)
            units = getattr(data_info.source, sasname + '_unit')
            if value is not None:
                source_entry[nxname] = _h5_float(value)
                source_entry[nxname].attrs['units'] = units
        _write_h5_vector(source_entry, data_info.source.beam_size,
            names=['beam_size_x', 'beam_size_y'],
            units=data_info.source.beam_size_unit, write_fn=_write_h5_float)

        # Collimation metadata
        if len(data_info.collimation) > 0:
            for i, coll_info in enumerate(data_info.collimation):
                collimation_entry = instrument_entry.create_group(
                    'sascollimation{0:0=2d}'.format(i + 1))
                collimation_entry.attrs['canSAS_class'] = 'SAScollimation'
                if coll_info.length is not None:
                    _write_h5_float(collimation_entry, coll_info.length, 'SDD')
                    collimation_entry['SDD'].attrs['units'] =\
                        coll_info.length_unit
                if coll_info.name is not None:
                    collimation_entry['name'] = _h5_string(coll_info.name)
        else:
            # Create a blank one - at least 1 collimation required by format
            instrument_entry.create_group('sascollimation01')

        # Detector metadata
        if len(data_info.detector) > 0:
            i = 1
            for i, det_info in enumerate(data_info.detector):
                detector_entry = instrument_entry.create_group(
                    'sasdetector{0:0=2d}'.format(i + 1))
                detector_entry.attrs['canSAS_class'] = 'SASdetector'
                if det_info.distance is not None:
                    _write_h5_float(detector_entry, det_info.distance, 'SDD')
                    detector_entry['SDD'].attrs['units'] =\
                        det_info.distance_unit
                if det_info.name is not None:
                    detector_entry['name'] = _h5_string(det_info.name)
                else:
                    detector_entry['name'] = _h5_string('')
                if det_info.slit_length is not None:
                    _write_h5_float(detector_entry, det_info.slit_length,
                                    'slit_length')
                    detector_entry['slit_length'].attrs['units'] =\
                        det_info.slit_length_unit
                _write_h5_vector(detector_entry, det_info.offset)
                # NXcanSAS doesn't save information about pitch, only roll
                # and yaw. The _write_h5_vector method writes vector.y, but we
                # need to write vector.z for yaw
                det_info.orientation.y = det_info.orientation.z
                _write_h5_vector(detector_entry, det_info.orientation,
                    names=['polar_angle', 'azimuthal_angle'])
                _write_h5_vector(detector_entry, det_info.beam_center,
                    names=['beam_center_x', 'beam_center_y'],
                    write_fn=_write_h5_float, units=det_info.beam_center_unit)
                _write_h5_vector(detector_entry, det_info.pixel_size,
                    names=['x_pixel_size', 'y_pixel_size'],
                    write_fn=_write_h5_float, units=det_info.pixel_size_unit)
        else:
            # Create a blank one - at least 1 detector required by format
            detector_entry = instrument_entry.create_group('sasdetector01')
            detector_entry.attrs['canSAS_class'] = 'SASdetector'
            detector_entry.attrs['name'] = ''

        # Process meta data
        for i, process in enumerate(data_info.process):
            process_entry = sasentry.create_group('sasprocess{0:0=2d}'.format(
                i + 1))
            process_entry.attrs['canSAS_class'] = 'SASprocess'
            if process.name:
                name = _h5_string(process.name)
                process_entry.create_dataset('name', data=name)
            if process.date:
                date = _h5_string(process.date)
                process_entry.create_dataset('date', data=date)
            if process.description:
                desc = _h5_string(process.description)
                process_entry.create_dataset('description', data=desc)
            for j, term in enumerate(process.term):
                # Don't save empty terms
                if term:
                    h5_term = _h5_string(term)
                    process_entry.create_dataset('term{0:0=2d}'.format(
                        j + 1), data=h5_term)
            for j, note in enumerate(process.notes):
                # Don't save empty notes
                if note:
                    h5_note = _h5_string(note)
                    process_entry.create_dataset('note{0:0=2d}'.format(
                        j + 1), data=h5_note)

        # Transmission Spectrum
        for i, trans in enumerate(data_info.trans_spectrum):
            trans_entry = sasentry.create_group(
                'sastransmission_spectrum{0:0=2d}'.format(i + 1))
            trans_entry.attrs['canSAS_class'] = 'SAStransmission_spectrum'
            trans_entry.attrs['signal'] = 'T'
            trans_entry.attrs['T_axes'] = 'T'
            trans_entry.attrs['name'] = trans.name
            # Timestamp can be a string or a date-like object. Only write if set
            if trans.timestamp:
                trans_entry.attrs['timestamp'] = trans.timestamp
            transmission = trans_entry.create_dataset('T',
                                                      data=trans.transmission)
            transmission.attrs['unertainties'] = 'Tdev'
            trans_entry.create_dataset('Tdev',
                                       data=trans.transmission_deviation)
            trans_entry.create_dataset('lambda', data=trans.wavelength)

        if np.isscalar(data_info.notes) and data_info.notes:
            # Handle scalars that aren't empty arrays, None, '', etc.
            notes = [np.string_(data_info.notes)]
        elif len(data_info.notes) > 1:
            # Handle iterables that aren't empty
            notes = [np.string_(n) for n in data_info.notes]
        else:
            # Notes are required in NXcanSAS format
            notes = [np.string_('')]
        if notes is not None:
            for i, note in enumerate(notes):
                note_entry = sasentry.create_group('sasnote{0}'.format(i))
                note_entry.attrs['canSAS_class'] = 'SASnote'
                note_entry.create_dataset('SASnote', data=note)

        f.close()

    def _write_1d_data(self, data_obj, data_entry):
        """
        Writes the contents of a Data1D object to a SASdata h5py Group

        :param data_obj: A Data1D object to write to the file
        :param data_entry: A h5py Group object representing the SASdata
        """
        data_entry.attrs['signal'] = 'I'
        data_entry.attrs['I_axes'] = 'Q'
        data_entry.attrs['Q_indices'] = [0]
        q_entry = data_entry.create_dataset('Q', data=data_obj.x)
        q_entry.attrs['units'] = data_obj.x_unit
        i_entry = data_entry.create_dataset('I', data=data_obj.y)
        i_entry.attrs['units'] = data_obj.y_unit
        if data_obj.dy is not None:
            i_entry.attrs['uncertainties'] = 'Idev'
            i_dev_entry = data_entry.create_dataset('Idev', data=data_obj.dy)
            i_dev_entry.attrs['units'] = data_obj.y_unit
        if data_obj.dx is not None:
            q_entry.attrs['resolutions'] = 'dQ'
            dq_entry = data_entry.create_dataset('dQ', data=data_obj.dx)
            dq_entry.attrs['units'] = data_obj.x_unit
        elif data_obj.dxl is not None:
            q_entry.attrs['resolutions'] = ['dQl','dQw']
            dql_entry = data_entry.create_dataset('dQl', data=data_obj.dxl)
            dql_entry.attrs['units'] = data_obj.x_unit
            dqw_entry = data_entry.create_dataset('dQw', data=data_obj.dxw)
            dqw_entry.attrs['units'] = data_obj.x_unit

    def _write_2d_data(self, data, data_entry):
        """
        Writes the contents of a Data2D object to a SASdata h5py Group

        :param data: A Data2D object to write to the file
        :param data_entry: A h5py Group object representing the SASdata
        """
        data_entry.attrs['signal'] = 'I'
        data_entry.attrs['I_axes'] = 'Qx,Qy'
        data_entry.attrs['Q_indices'] = [0,1]

        (n_rows, n_cols) = (len(data.y_bins), len(data.x_bins))

        if (n_rows == 0 and n_cols == 0) or (n_cols*n_rows != data.data.size):
            # Calculate rows and columns, assuming detector is square
            # Same logic as used in PlotPanel.py _get_bins
            n_cols = int(np.floor(np.sqrt(len(data.qy_data))))
            n_rows = int(np.floor(len(data.qy_data) / n_cols))

            if n_rows * n_cols != len(data.qy_data):
                raise ValueError("Unable to calculate dimensions of 2D data")

        intensity = np.reshape(data.data, (n_rows, n_cols))
        qx = np.reshape(data.qx_data, (n_rows, n_cols))
        qy = np.reshape(data.qy_data, (n_rows, n_cols))

        i_entry = data_entry.create_dataset('I', data=intensity)
        i_entry.attrs['units'] = data.I_unit
        qx_entry = data_entry.create_dataset('Qx', data=qx)
        qx_entry.attrs['units'] = data.Q_unit
        qy_entry = data_entry.create_dataset('Qy', data=qy)
        qy_entry.attrs['units'] = data.Q_unit
        if (data.err_data is not None
                and not all(v is None for v in data.err_data)):
            d_i = np.reshape(data.err_data, (n_rows, n_cols))
            i_entry.attrs['uncertainties'] = 'Idev'
            i_dev_entry = data_entry.create_dataset('Idev', data=d_i)
            i_dev_entry.attrs['units'] = data.I_unit
        if (data.dqx_data is not None
                and not all(v is None for v in data.dqx_data)):
            qx_entry.attrs['resolutions'] = 'dQx'
            dqx_entry = data_entry.create_dataset('dQx', data=data.dqx_data)
            dqx_entry.attrs['units'] = data.Q_unit
        if (data.dqy_data is not None
                and not all(v is None for v in data.dqy_data)):
            qy_entry.attrs['resolutions'] = 'dQy'
            dqy_entry = data_entry.create_dataset('dQy', data=data.dqy_data)
            dqy_entry.attrs['units'] = data.Q_unit
        if data.mask is not None and not all(v is None for v in data.mask):
            data_entry.attrs['mask'] = "mask"
            mask = np.invert(np.asarray(data.mask, dtype=bool))
            data_entry.create_dataset('mask', data=mask)



"""
    NXcanSAS data reader for reading HDF5 formatted CanSAS files.
"""

import logging
import h5py
import numpy as np
import re
import os
import traceback
from typing import Any, Union, Optional

from sasdata.dataloader.data_info import plottable_1D, plottable_2D, Data1D, Data2D, DataInfo, Process, Aperture,\
    Collimation, TransmissionSpectrum, Detector
from sasdata.data_util.loader_exceptions import FileContentsException, DefaultReaderException
from sasdata.dataloader.filereader import FileReader, decode

logger = logging.getLogger(__name__)


def h5attr(node, key, default=None):
    value = node.attrs.get(key, default)
    if isinstance(value, np.ndarray) and value.dtype.char == 'S':
        return [decode(el) for el in value]
    elif isinstance(value, list):
        return [decode(el) for el in value]
    else:
        return decode(value)


class Reader(FileReader):
    """
    A class for reading in NXcanSAS data files. The current implementation has
    been tested to load data generated by multiple facilities, all of which are
    known to produce NXcanSAS standards compliant data. Any number of data sets
    may be present within the file and any dimensionality of data may be used.
    Currently 1D and 2D SAS data sets are supported, but should be immediately
    extensible to SESANS data.

    Any number of SASdata groups  may be present in a SASentry and the data
    within each SASdata group can be a single 1D I(Q), multi-framed 1D I(Q),
    2D I(Qx, Qy) or multi-framed 2D I(Qx, Qy).

    :Dependencies:
        The NXcanSAS HDF5 reader requires h5py => v2.5.0 or later.
    """

    # CanSAS version
    cansas_version = 2.0
    # Data type name
    type_name = "NXcanSAS"
    # Wildcards
    type = ["NXcanSAS HDF5 Files (*.h5)|*.h5|"]
    # List of allowed extensions
    ext = ['.h5', '.H5']
    # Flag to bypass extension check
    allow_all = True

    def get_file_contents(self):
        """
        This is the general read method that all SasView data_loaders must have.

        :param filename: A path for an HDF5 formatted CanSAS 2D data file.
        :return: List of Data1D/2D objects and/or a list of errors.
        """
        # Reinitialize when loading a new data file to reset all class variables
        self.reset_state()

        filename = self.f_open.name
        self.f_open.close() # IO handled by h5py

        # Check that the file exists
        if os.path.isfile(filename):
            basename = os.path.basename(filename)
            _, extension = os.path.splitext(basename)
            # If the file type is not allowed, return empty list
            if extension in self.ext or self.allow_all:
                # Load the data file
                try:
                    self.raw_data = h5py.File(filename, 'r')
                except Exception as exc:
                    if extension not in self.ext:
                        msg = "NXcanSAS Reader could not load file {}".format(
                            basename + extension)
                        raise DefaultReaderException(msg)
                    raise FileContentsException(exc)
                try:
                    # Read in all child elements of top level SASroot
                    self.read_children(self.raw_data, [])
                    # Add the last data set to the list of outputs
                    self.add_data_set()
                except Exception as exc:
                    raise FileContentsException(exc)
                finally:
                    # Close the data file
                    self.raw_data.close()

                for data_set in self.output:
                    if isinstance(data_set, Data1D):
                        if data_set.x.size < 5:
                            exception = FileContentsException(
                                "Fewer than 5 data points found.")
                            data_set.errors.append(exception)

    def reset_state(self):
        """
        Create the reader object and define initial states for class variables
        """
        super(Reader, self).reset_state()
        self.data1d = []
        self.data2d = []
        self.raw_data = None
        self.multi_frame = False
        self.data_frames = []
        self.data_uncertainty_frames = []
        self.errors = []
        self.logging = []
        self.q_names = []
        self.mask_name = u''
        self.i_name = u''
        self.i_node = u''
        self.i_uncertainties_name = u''
        self.q_uncertainty_names = []
        self.q_resolution_names = []
        self.parent_class = u''
        self.detector = Detector()
        self.collimation = Collimation()
        self.aperture = Aperture()
        self.process = Process()
        self.trans_spectrum = TransmissionSpectrum()

    def read_children(self, data: dict, parent_list: list):
        """
        A recursive method for stepping through the hierarchical data file.

        :param data: h5py Group object of any kind
        :param parent: h5py Group parent name
        """

        # Loop through each element of the parent and process accordingly
        for key in data.keys():
            # Get all information for the current key
            value = data.get(key)
            class_name = h5attr(value, u'canSAS_class')
            if isinstance(class_name, (list, tuple, np.ndarray)):
                class_name = class_name[0]
            if class_name is None:
                class_name = h5attr(value, u'NX_class')
            if class_name is not None:
                class_prog = re.compile(class_name)
            else:
                class_prog = re.compile(value.name)

            if isinstance(value, h5py.Group):
                # Set parent class before recursion
                last_parent_class = self.parent_class
                self.parent_class = class_name
                parent_list.append(key)
                # If a new sasentry, store the current data sets and create
                # a fresh Data1D/2D object
                if class_prog.match(u'SASentry'):
                    self.add_data_set()
                elif class_prog.match(u'SASdata'):
                    self._find_data_attributes(value)
                    self._initialize_new_data_set(value)
                # Recursion step to access data within the group
                try:
                    self.read_children(value, parent_list)
                    self.add_intermediate()
                except Exception as e:
                    self.current_datainfo.errors.append(str(e))
                    logger.debug(traceback.format_exc())
                # Reset parent class when returning from recursive method
                self.parent_class = last_parent_class
                parent_list.remove(key)

            elif isinstance(value, h5py.Dataset):
                # If this is a dataset, store the data appropriately
                data_set = value[()]
                unit = self._get_unit(value)
                # Put scalars into lists to be sure they are iterable
                if np.isscalar(data_set):
                    data_set = [data_set]

                for data_point in data_set:
                    if isinstance(data_point, np.ndarray):
                        if data_point.dtype.char == 'S':
                            data_point = decode(bytes(data_point))
                    else:
                        data_point = decode(data_point)
                    # Top Level Meta Data
                    if key == u'definition':
                        if isinstance(data_set, str):
                            self.current_datainfo.meta_data['reader'] = data_set
                            break
                        else:
                            self.current_datainfo.meta_data[
                                'reader'] = data_point
                    # Run
                    elif key == u'run':
                        try:
                            run_name = h5attr(value, 'name', default='name')
                            run_dict = {data_point: run_name}
                            self.current_datainfo.run_name = run_dict
                        except Exception:
                            pass
                        if isinstance(data_set, str):
                            self.current_datainfo.run.append(data_set)
                            break
                        else:
                            self.current_datainfo.run.append(data_point)
                    # Title
                    elif key == u'title':
                        if isinstance(data_set, str):
                            self.current_datainfo.title = data_set
                            break
                        else:
                            self.current_datainfo.title = data_point
                    # Note
                    elif key == u'SASnote':
                        self.current_datainfo.notes.append(data_set)
                        break
                    # Sample Information
                    elif self.parent_class == u'SASsample':
                        self.process_sample(data_point, key)
                    # Instrumental Information
                    elif (key == u'name'
                          and self.parent_class == u'SASinstrument'):
                        self.current_datainfo.instrument = data_point
                    # Detector
                    elif self.parent_class == u'SASdetector':
                        self.process_detector(data_point, key, unit)
                    # Collimation
                    elif self.parent_class == u'SAScollimation':
                        self.process_collimation(data_point, key, unit)
                    # Aperture
                    elif self.parent_class == u'SASaperture':
                        self.process_aperture(data_point, key)
                    # Process Information
                    elif self.parent_class == u'SASprocess': # CanSAS 2.0
                        self.process_process(data_point, key)
                    # Source
                    elif self.parent_class == u'SASsource':
                        self.process_source(data_point, key, unit)
                    # Everything else goes in meta_data
                    elif self.parent_class == u'SASdata':
                        if isinstance(self.current_dataset, plottable_2D):
                            self.process_2d_data_object(data_set, key, unit)
                        else:
                            self.process_1d_data_object(data_set, key, unit)

                        break
                    elif self.parent_class == u'SAStransmission_spectrum':
                        self.process_trans_spectrum(data_set, key)
                        break
                    else:
                        new_key = self._create_unique_key(
                            self.current_datainfo.meta_data, key)
                        self.current_datainfo.meta_data[new_key] = data_point

            else:
                # I don't know if this reachable code
                self.errors.append("ShouldNeverHappenException")

    def process_1d_data_object(self, data_set: np.array, key: str, unit: str):
        """
        SASdata processor method for 1d data items
        :param data_set: data from HDF5 file
        :param key: canSAS_class attribute
        :param unit: unit attribute
        """
        if key == self.i_name:
            if self.multi_frame:
                for x in range(0, data_set.shape[0]):
                    self.data_frames.append(data_set[x].flatten())
            else:
                self.current_dataset.y = data_set.flatten()
                self.current_dataset.yaxis("Intensity", unit)
        elif key == self.i_uncertainties_name:
            if self.multi_frame:
                for x in range(0, data_set.shape[0]):
                    self.data_uncertainty_frames.append(data_set[x].flatten())
            self.current_dataset.dy = data_set.flatten()
        elif key in self.q_names:
            self.current_dataset.xaxis("Q", unit)
            self.current_dataset.x = data_set.flatten()
        elif key in self.q_resolution_names:
            if (len(self.q_resolution_names) > 1
                    and np.where(self.q_resolution_names == key)[0] == 0):
                self.current_dataset.dxw = data_set.flatten()
            elif (len(self.q_resolution_names) > 1
                  and np.where(self.q_resolution_names == key)[0] == 1):
                self.current_dataset.dxl = data_set.flatten()
            else:
                self.current_dataset.dx = data_set.flatten()
        elif key in self.q_uncertainty_names:
            if (len(self.q_uncertainty_names) > 1
                    and np.where(self.q_uncertainty_names == key)[0] == 0):
                self.current_dataset.dxw = data_set.flatten()
            elif (len(self.q_uncertainty_names) > 1
                  and np.where(self.q_uncertainty_names == key)[0] == 1):
                self.current_dataset.dxl = data_set.flatten()
            else:
                self.current_dataset.dx = data_set.flatten()
        elif key == self.mask_name:
            self.current_dataset.mask = data_set.flatten()
        elif key == u'wavelength':
            self.current_datainfo.source.wavelength = data_set[0]
            self.current_datainfo.source.wavelength_unit = unit

    def process_2d_data_object(self, data_set: np.array, key: str, unit: str):
        if key == self.i_name:
            self.current_dataset.data = data_set
            self.current_dataset.zaxis("Intensity", unit)
        elif key == self.i_uncertainties_name:
            self.current_dataset.err_data = data_set.flatten()
        elif key in self.q_names:
            self.current_dataset.xaxis("Q_x", unit)
            self.current_dataset.yaxis("Q_y", unit)
            if self.q_names[0] == self.q_names[1]:
                # All q data in a single array
                self.current_dataset.qx_data = data_set[0]
                self.current_dataset.qy_data = data_set[1]
            elif self.q_names.index(key) == 0:
                self.current_dataset.qx_data = data_set
            elif self.q_names.index(key) == 1:
                self.current_dataset.qy_data = data_set
        elif key in self.q_resolution_names:
            if (len(self.q_resolution_names) == 1
                    or (self.q_resolution_names[0]
                        == self.q_resolution_names[1])):
                self.current_dataset.dqx_data = data_set[0].flatten()
                self.current_dataset.dqy_data = data_set[1].flatten()
            elif self.q_resolution_names[0] == key:
                self.current_dataset.dqx_data = data_set.flatten()
            else:
                self.current_dataset.dqy_data = data_set.flatten()
        elif key in self.q_uncertainty_names:
            if (len(self.q_uncertainty_names) == 1
                    or (self.q_uncertainty_names[0]
                        == self.q_uncertainty_names[1])):
                self.current_dataset.dqx_data = data_set[0].flatten()
                self.current_dataset.dqy_data = data_set[1].flatten()
            elif self.q_uncertainty_names[0] == key:
                self.current_dataset.dqx_data = data_set.flatten()
            else:
                self.current_dataset.dqy_data = data_set.flatten()
        elif key == self.mask_name:
            self.current_dataset.mask = data_set.flatten()
        elif key == u'Qy':
            self.current_dataset.yaxis("Q_y", unit)
            self.current_dataset.qy_data = data_set.flatten()
        elif key == u'Qydev':
            self.current_dataset.dqy_data = data_set.flatten()
        elif key == u'Qx':
            self.current_dataset.xaxis("Q_x", unit)
            self.current_dataset.qx_data = data_set.flatten()
        elif key == u'Qxdev':
            self.current_dataset.dqx_data = data_set.flatten()

    def process_trans_spectrum(self, data_set: np.array, key: str):
        """
        SAStransmission_spectrum processor
        :param data_set: data from HDF5 file
        :param key: canSAS_class attribute
        """
        if key == u'T':
            self.trans_spectrum.transmission = data_set.flatten()
        elif key == u'Tdev':
            self.trans_spectrum.transmission_deviation = data_set.flatten()
        elif key == u'lambda':
            self.trans_spectrum.wavelength = data_set.flatten()

    def process_sample(self, data_point: Any, key: str):
        """
        SASsample processor
        :param data_point: Single point from an HDF5 data file
        :param key: class name data_point was taken from
        """
        if key == u'Title':
            self.current_datainfo.sample.name = data_point
        elif key == u'name':
            self.current_datainfo.sample.name = data_point
        elif key == u'ID':
            self.current_datainfo.sample.name = data_point
        elif key == u'thickness':
            self.current_datainfo.sample.thickness = data_point
        elif key == u'temperature':
            self.current_datainfo.sample.temperature = data_point
        elif key == u'transmission':
            self.current_datainfo.sample.transmission = data_point
        elif key == u'x_position':
            self.current_datainfo.sample.position.x = data_point
        elif key == u'y_position':
            self.current_datainfo.sample.position.y = data_point
        elif key == u'pitch':
            self.current_datainfo.sample.orientation.x = data_point
        elif key == u'yaw':
            self.current_datainfo.sample.orientation.y = data_point
        elif key == u'roll':
            self.current_datainfo.sample.orientation.z = data_point
        elif key == u'details':
            self.current_datainfo.sample.details.append(data_point)

    def process_detector(self, data_point: Any, key: str, unit: str):
        """
        SASdetector processor
        :param data_point: Single point from an HDF5 data file
        :param key: class name data_point was taken from
        :param unit: unit attribute from data set
        """
        if key == u'name':
            self.detector.name = data_point
        elif key == u'SDD':
            self.detector.distance = float(data_point)
            self.detector.distance_unit = unit
        elif key == u'slit_length':
            self.detector.slit_length = float(data_point)
            self.detector.slit_length_unit = unit
        elif key == u'x_position':
            self.detector.offset.x = float(data_point)
            self.detector.offset_unit = unit
        elif key == u'y_position':
            self.detector.offset.y = float(data_point)
            self.detector.offset_unit = unit
        elif key == u'pitch':
            self.detector.orientation.x = float(data_point)
            self.detector.orientation_unit = unit
        elif key == u'roll':
            self.detector.orientation.z = float(data_point)
            self.detector.orientation_unit = unit
        elif key == u'yaw':
            self.detector.orientation.y = float(data_point)
            self.detector.orientation_unit = unit
        elif key == u'beam_center_x':
            self.detector.beam_center.x = float(data_point)
            self.detector.beam_center_unit = unit
        elif key == u'beam_center_y':
            self.detector.beam_center.y = float(data_point)
            self.detector.beam_center_unit = unit
        elif key == u'x_pixel_size':
            self.detector.pixel_size.x = float(data_point)
            self.detector.pixel_size_unit = unit
        elif key == u'y_pixel_size':
            self.detector.pixel_size.y = float(data_point)
            self.detector.pixel_size_unit = unit

    def process_collimation(self, data_point: Any, key: str, unit: str):
        """
        SAScollimation processor
        :param data_point: Single point from an HDF5 data file
        :param key: class name data_point was taken from
        :param unit: unit attribute from data set
        """
        if key == u'distance':
            self.collimation.length = data_point
            self.collimation.length_unit = unit
        elif key == u'name':
            self.collimation.name = data_point

    def process_aperture(self, data_point: Any, key: str):
        """
        SASaperture processor
        :param data_point: Single point from an HDF5 data file
        :param key: class name data_point was taken from
        """
        if key == u'shape':
            self.aperture.shape = data_point
        elif key == u'x_gap':
            self.aperture.size.x = data_point
        elif key == u'y_gap':
            self.aperture.size.y = data_point

    def process_source(self, data_point: Any, key: str, unit: str):
        """
        SASsource processor
        :param data_point: Single point from an HDF5 data file
        :param key: class name data_point was taken from
        :param unit: unit attribute from data set
        """
        if key == u'incident_wavelength':
            self.current_datainfo.source.wavelength = data_point
            self.current_datainfo.source.wavelength_unit = unit
        elif key == u'wavelength_max':
            self.current_datainfo.source.wavelength_max = data_point
            self.current_datainfo.source.wavelength_max_unit = unit
        elif key == u'wavelength_min':
            self.current_datainfo.source.wavelength_min = data_point
            self.current_datainfo.source.wavelength_min_unit = unit
        elif key == u'incident_wavelength_spread':
            self.current_datainfo.source.wavelength_spread = data_point
            self.current_datainfo.source.wavelength_spread_unit = unit
        elif key == u'beam_size_x':
            self.current_datainfo.source.beam_size.x = data_point
            self.current_datainfo.source.beam_size_unit = unit
        elif key == u'beam_size_y':
            self.current_datainfo.source.beam_size.y = data_point
            self.current_datainfo.source.beam_size_unit = unit
        elif key == u'beam_shape':
            self.current_datainfo.source.beam_shape = data_point
        elif key == u'radiation':
            self.current_datainfo.source.radiation = data_point
        elif key == u'type':
            self.current_datainfo.source.type = data_point
        elif key == u'probe':
            self.current_datainfo.source.probe = data_point

    def process_process(self, data_point: Any, key: str):
        """
        SASprocess processor
        :param data_point: Single point from an HDF5 data file
        :param key: class name data_point was taken from
        """
        term_match = re.compile(u'^term[0-9]+$')
        if key == u'Title':  # CanSAS 2.0
            self.process.name = data_point
        elif key == u'name':  # NXcanSAS
            self.process.name = data_point
        elif key == u'description':
            self.process.description = data_point
        elif key == u'date':
            self.process.date = data_point
        elif term_match.match(key):
            self.process.term.append(data_point)
        else:
            self.process.notes.append(data_point)

    def add_intermediate(self):
        """
        This method stores any intermediate objects within the final data set
        after fully reading the set.

        :param parent: The NXclass name for the h5py Group object that just
                       finished being processed
        """

        if self.parent_class == u'SASprocess':
            self.current_datainfo.process.append(self.process)
            self.process = Process()
        elif self.parent_class == u'SASdetector':
            self.current_datainfo.detector.append(self.detector)
            self.detector = Detector()
        elif self.parent_class == u'SAStransmission_spectrum':
            self.current_datainfo.trans_spectrum.append(self.trans_spectrum)
            self.trans_spectrum = TransmissionSpectrum()
        elif self.parent_class == u'SAScollimation':
            self.current_datainfo.collimation.append(self.collimation)
            self.collimation = Collimation()
        elif self.parent_class == u'SASaperture':
            self.collimation.aperture.append(self.aperture)
            self.aperture = Aperture()
        elif self.parent_class == u'SASdata':
            if isinstance(self.current_dataset, plottable_2D):
                self.data2d.append(self.current_dataset)
            elif isinstance(self.current_dataset, plottable_1D):
                if self.multi_frame:
                    for x in range(0, len(self.data_frames)):
                        self.current_dataset.y = self.data_frames[x]
                        if len(self.data_uncertainty_frames) > x:
                            self.current_dataset.dy = \
                                self.data_uncertainty_frames[x]
                        self.data1d.append(self.current_dataset)
                else:
                    self.data1d.append(self.current_dataset)

    def final_data_cleanup(self):
        """
        Does some final cleanup and formatting on self.current_datainfo and
        all data1D and data2D objects and then combines the data and info into
        Data1D and Data2D objects
        """
        # Type cast data arrays to float64
        if len(self.current_datainfo.trans_spectrum) > 0:
            spectrum_list = []
            for spectrum in self.current_datainfo.trans_spectrum:
                spectrum.transmission = spectrum.transmission.astype(np.float64)
                spectrum.transmission_deviation = \
                    spectrum.transmission_deviation.astype(np.float64)
                spectrum.wavelength = spectrum.wavelength.astype(np.float64)
                if len(spectrum.transmission) > 0:
                    spectrum_list.append(spectrum)
            self.current_datainfo.trans_spectrum = spectrum_list

        # Append errors to dataset and reset class errors
        self.current_datainfo.errors = self.errors
        self.errors = []

        # Combine all plottables with datainfo and append each to output
        # Type cast data arrays to float64 and find min/max as appropriate
        for dataset in self.data2d:
            # Calculate the actual Q matrix
            try:
                if dataset.q_data.size <= 1:
                    dataset.q_data = np.sqrt(dataset.qx_data
                                             * dataset.qx_data
                                             + dataset.qy_data
                                             * dataset.qy_data).flatten()
            except:
                dataset.q_data = None

            if dataset.data.ndim == 2:
                dataset.y_bins = np.unique(dataset.qy_data.flatten())
                dataset.x_bins = np.unique(dataset.qx_data.flatten())
                dataset.data = dataset.data.flatten()
                dataset.qx_data = dataset.qx_data.flatten()
                dataset.qy_data = dataset.qy_data.flatten()

            try:
                iter(dataset.mask)
                dataset.mask = np.invert(np.asarray(dataset.mask, dtype=bool))
            except TypeError:
                dataset.mask = np.ones(dataset.data.shape, dtype=bool)
            self.current_dataset = dataset
            self.send_to_output()

        for dataset in self.data1d:
            self.current_dataset = dataset
            self.send_to_output()

    def add_data_set(self):
        """
        Adds the current_dataset to the list of outputs after preforming final
        processing on the data and then calls a private method to generate a
        new data set.
        """

        if self.current_datainfo and self.current_dataset:
            self.final_data_cleanup()
        self.data_frames = []
        self.data_uncertainty_frames = []
        self.data1d = []
        self.data2d = []
        self.current_datainfo = DataInfo()

    def _initialize_new_data_set(self, value=None):
        """
        A private class method to generate a new 1D or 2D data object based on
        the type of data within the set. Outside methods should call
        add_data_set() to be sure any existing data is stored properly.

        :param parent_list: List of names of parent elements
        """
        if self._is_2d_not_multi_frame(value):
            self.current_dataset = plottable_2D()
        else:
            x = np.array(0)
            y = np.array(0)
            self.current_dataset = plottable_1D(x, y)
        self.current_datainfo.filename = self.raw_data.filename

    @staticmethod
    def as_list_or_array(data: Any) -> Union[list, np.ndarray]:
        """
        Return value as a list if not already a list or array.
        :param iterable:
        :return:
        """
        if not (isinstance(data, np.ndarray) or isinstance(data, list)):
            data = data.split(",") if isinstance(data, str) else [data]
        return data

    def _find_data_attributes(self, value):
        """
        A class to find the indices for Q, the name of the Qdev and Idev, and
        the name of the mask.
        :param value: SASdata/NXdata HDF5 Group
        """
        # Initialize values to base types
        self.mask_name = u''
        self.i_name = u''
        self.i_node = u''
        self.i_uncertainties_name = u''
        self.q_names = []
        self.q_uncertainty_names = []
        self.q_resolution_names = []
        # Get attributes
        signal = h5attr(value, "signal", "I")
        i_axes = h5attr(value, "I_axes", ["Q"])
        q_indices = h5attr(value, "Q_indices", [0])
        i_axes = self.as_list_or_array(i_axes)
        keys = value.keys()
        # Assign attributes to appropriate class variables
        self.q_names = [i_axes[int(v)] for v in self.as_list_or_array(q_indices)]
        self.mask_name = h5attr(value, "mask")
        self.i_name = signal
        self.i_node = value.get(self.i_name)
        for item in self.q_names:
            if item in keys:
                q_vals = value.get(item)
                uncertainties = h5attr(q_vals, "uncertainties")
                if uncertainties is None:
                    uncertainties = h5attr(q_vals, "uncertainty")
                if isinstance(uncertainties, str):
                    uncertainties = uncertainties.split(",")
                if uncertainties is not None:
                    self.q_uncertainty_names = uncertainties
                resolutions = h5attr(q_vals, "resolutions")
                if isinstance(resolutions, str):
                    resolutions = resolutions.split(",")
                if resolutions is not None:
                    self.q_resolution_names = resolutions
        if self.i_name in keys:
            i_vals = value.get(self.i_name)
            self.i_uncertainties_name = h5attr(i_vals, "uncertainties")
            if self.i_uncertainties_name is None:
                self.i_uncertainties_name = h5attr(i_vals, "uncertainty")

    def _is_2d_not_multi_frame(self, value, i_base="", q_base=""):
        """
        A private class to determine if the data set is 1d or 2d.

        :param value: Nexus/NXcanSAS data group
        :param basename: Approximate name of an entry to search for
        :return: True if 2D, otherwise false
        """
        i_basename = i_base if i_base != "" else self.i_name
        i_vals = value.get(i_basename)
        q_basename = q_base if q_base != "" else self.q_names
        q_vals = value.get(q_basename[0])
        self.multi_frame = (i_vals is not None and q_vals is not None
                            and len(i_vals.shape) != 1
                            and len(q_vals.shape) == 1)
        return (i_vals is not None and len(i_vals.shape) != 1
                and not self.multi_frame)

    def _create_unique_key(self, dictionary: dict, name: str, numb: Optional[int] = 0) -> str:
        """
        Create a unique key value for any dictionary to prevent overwriting
        Recurses until a unique key value is found.

        :param dictionary: A dictionary with any number of entries
        :param name: The index of the item to be added to dictionary
        :param numb: The number to be appended to the name, starts at 0
        :return: The new name for the dictionary entry
        """
        if dictionary.get(name) is not None:
            numb += 1
            name = name.split("_")[0]
            name += "_{0}".format(numb)
            name = self._create_unique_key(dictionary, name, numb)
        return name

    def _get_unit(self, value: h5py.Dataset) -> str:
        """
        Find the unit for a particular value within the h5py dictionary

        :param value: attribute dictionary for a particular value set
        :return: unit for the value passed to the method
        """
        unit = h5attr(value, u'units')
        if unit is None:
            unit = h5attr(value, u'unit')
        return unit

    def write(self, filename: str, dataset: Union[Data1D, Data2D]):
        """
        Export data in NXcanSAS format

        :param filename: File path where the data will be saved
        :param dataset: DataInfo object that will be converted to NXcanSAS
        :return: None
        """
        from sasdata.file_converter.nxcansas_writer import NXcanSASWriter
        writer = NXcanSASWriter()
        writer.write(dataset=[dataset], filename=filename)
